#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Paul Rougieux and Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

JRC biomass Project.
Unit D1 Bioeconomy.

You can use this object at the ipython console with the following examples.
"""
# First party modules
import logging

# Third party modules
from sqlalchemy import Integer, Float, SmallInteger, Text, UniqueConstraint
from sqlalchemy import Table, Column, MetaData, or_
from sqlalchemy import create_engine, inspect, select
from sqlalchemy.sql import func
from sqlalchemy.schema import CreateSchema
from sqlalchemy_utils import database_exists, create_database
import pandas

# Internal modules
from biotrade import data_dir, database_url
from biotrade.common.database import Database


class DatabaseFaostat(Database):
    """
    Database to store UN Comtrade data.

    Select forestry production data for Italy

        >>> from biotrade.faostat import faostat
        >>> db = faostat.db
        >>> fp_ita = db.select(table="forestry_production",
        >>>                    reporter=["Italy"])

    More examples are available under the documentation of the `select` method below.

    Select data for Italy using an SQL Alchemy select statement. Return results
    using a pandas data frame:

        >>> import pandas
        >>> from biotrade.faostat import faostat
        >>> db = faostat.db
        >>> reporter = db.forestry_production.columns.get("reporter")
        >>> statement = db.forestry_production.select().where(reporter == "Italy")
        >>> with faostat.db.engine.connect() as conn:
        >>>     fp_it = pandas.read_sql_query(statement, conn)

    Use a SQL select statement directly

        >>> from sqlalchemy import text
        >>> # In case you are using an SQLite database
        >>> schema = "main"
        >>> # In case you are using a PostGreSQL database
        >>> schema = "raw_faostat"
        >>> query = text(f"SELECT * FROM {schema}.forestry_production WHERE reporter = 'Italy'")
        >>> with faostat.db.engine.connect() as conn:
        >>>     fp_it = pandas.read_sql_query(query, conn)
        >>> query = text(f"SELECT * FROM {schema}.forestry_trade  WHERE reporter = 'Italy'")
        >>> with faostat.db.engine.connect() as conn:
        >>>     ft_it = pandas.read_sql_query(query, conn)

    Database update methods are in the faostat.pump object.

    """

    # To be overwritten by the children
    database_url = None
    schema = None

    # Log debug and error messages
    logger = logging.getLogger("biotrade.comtrade")

    def __init__(self, parent):
        # Default attributes #
        self.parent = parent
        # Database configuration
        self.engine = create_engine(self.database_url)
        # Create the database if it doesn't exist
        if not database_exists(self.engine.url):
            create_database(self.engine.url)

        # SQL Alchemy metadata
        self.metadata = MetaData(schema=self.schema)
        self.metadata.bind = self.engine
        self.inspector = inspect(self.engine)

        # Create the schema if it doesn't exist.
        # Exclude the SQLite engine because there is only a default schemas for that engine.
        # And the SQLite dialect doesn't have a has_schema() method.
        if hasattr(self.engine.dialect, "has_schema") and callable(
            getattr(self.engine.dialect, "has_schema")
        ):
            with self.engine.connect() as conn:
                if not self.engine.dialect.has_schema(conn, self.schema):
                    conn.execute(CreateSchema(self.schema))
                    conn.commit()

        # Describe table metadata
        self.forestry_production = self.describe_production_table(
            name="forestry_production"
        )
        self.country = self.describe_country_table(name="country")
        self.crop_production = self.describe_production_table(name="crop_production")
        self.crop_trade = self.describe_trade_table(name="crop_trade")
        self.food_balance = self.describe_food_balance_table(name="food_balance")
        self.forest_land = self.describe_land_table(name="forest_land")
        self.forestry_trade = self.describe_trade_table(name="forestry_trade")
        self.land_cover = self.describe_land_table(name="land_cover")
        self.land_use = self.describe_land_table(name="land_use")
        self.tables = {
            "country": self.country,
            "crop_production": self.crop_production,
            "crop_trade": self.crop_trade,
            "food_balance": self.food_balance,
            "forest_land": self.forest_land,
            "forestry_production": self.forestry_production,
            "forestry_trade": self.forestry_trade,
            "land_cover": self.land_cover,
            "land_use": self.land_use,
        }
        # Create tables if they don't exist
        for table in self.tables.values():
            self.create_if_not_existing(table)

    def describe_production_table(self, name):
        """Define the metadata of a table containing production data.

        The unique constraint is a very important part of the table structure.
        It makes sure that there will be no duplicated flows.

        Note the "product" column is left empty, removed from the data frame
        before it is stored in the database because it would be too large. The
        text description of a product (commodity) is available in the products table.

        Experiment on database size test: dropping the following text columns from the definition

             Column("reporter", Text),
             Column("product", Text),
             Column("element", Text),
             Column("unit", Text),
             Column("flag", Text),

        And dropping these columns from the data frame before loading it back to the database

            >>> from biotrade.faostat import faostat
            >>> fp = faostat.pump.read_df("forestry_production")
            >>> fp.drop(columns=["reporter", "product", "element", "unit", "flag"], inplace=True)
            >>> faostat.db_sqlite.append(fp, "forestry_production")

        Leads to a size reduction from 250 Mb to 99 Mb

            cd ~/repos/biotrade_data/faostat/ && ls -l faostat.db
            -rw-r--r-- 1 paul paul 99246080 Nov  2 05:34 faostat.db
            -rw-r--r-- 1 paul paul 251060224 Nov  2 05:43 faostat.db

        As a result of this little experiment, it is probably not worth using
        product and country tables with joins on indexes, because the gain will
        not be very large.
        """
        table = Table(
            name,
            self.metadata,
            Column("reporter_code", Integer),
            Column("reporter", Text),
            Column("product_code", Integer),
            Column("product", Text),
            Column("element_code", Integer),
            Column("element", Text),
            Column("period", Integer),
            Column("year", Integer),
            Column("unit", Text),
            Column("value", Float),
            Column("flag", Text),
            UniqueConstraint(
                "period",
                "reporter_code",
                "product_code",
                "element_code",
                "unit",
                "flag",
            ),
            schema=self.schema,
        )
        return table

    def describe_trade_table(self, name):
        """Define the metadata of a table containing trade data."""
        table = Table(
            name,
            self.metadata,
            Column("reporter_code", Integer),
            Column("reporter", Text),
            Column("partner_code", Integer),
            Column("partner", Text),
            Column("product_code", Integer),
            Column("product", Text),
            Column("element_code", Integer),
            Column("element", Text),
            Column("period", Integer),
            Column("year", Integer),
            Column("unit", Text),
            Column("value", Float),
            Column("flag", Text),
            UniqueConstraint(
                "period",
                "reporter_code",
                "partner_code",
                "product_code",
                "element_code",
                "unit",
                "flag",
            ),
            schema=self.schema,
        )
        return table

    def describe_land_table(self, name):
        """Define the metadata of a table containing land use data."""
        table = Table(
            name,
            self.metadata,
            Column("reporter_code", Integer),
            Column("reporter", Text),
            Column("item_code", Integer),
            Column("item", Text),
            Column("element_code", Integer),
            Column("element", Text),
            Column("period", Integer),
            Column("year", Integer),
            Column("source_code", Integer),
            Column("source", Text),
            Column("unit", Text),
            Column("value", Float),
            Column("flag", Text),
            Column("note", Text),
            UniqueConstraint(
                "period",
                "reporter_code",
                "item_code",
                "element_code",
                "unit",
                "flag",
            ),
            schema=self.schema,
        )
        return table

    def describe_food_balance_table(self, name):
        """Define the metadata for the Supply and Utilisation Accounts table
        containing data about food balance, production and trade
        """
        table = Table(
            name,
            self.metadata,
            Column("reporter_code", Integer),
            Column("reporter", Text),
            Column("product_code", Text),
            Column("product", Text),
            Column("element_code", Integer),
            Column("element", Text),
            Column("period", Integer),
            Column("year", Integer),
            Column("unit", Text),
            Column("value", Float),
            Column("flag", Text),
            UniqueConstraint(
                "period",
                "reporter_code",
                "product_code",
                "element_code",
                "unit",
                "flag",
            ),
            schema=self.schema,
        )
        return table

    def describe_country_table(self, name):
        """Define the metadata of a table containing country data."""
        table = Table(
            name,
            self.metadata,
            Column("continent_code", Integer),
            Column("continent", Text),
            Column("sub_continent_code", Integer),
            Column("sub_continent", Text),
            Column("eu27", SmallInteger),
            Column("country_code", Integer),
            Column("country_name", Text),
            UniqueConstraint(
                "country_code",
            ),
            schema=self.schema,
        )
        return table

    def read_sql_query(self, stmt):
        """A wrapper around pandas.read_sql_query"""
        with self.engine.connect() as conn:
            df = pandas.read_sql_query(stmt, conn)
        return df

    def select(
        self,
        table,
        reporter=None,
        partner=None,
        product=None,
        element=None,
        reporter_code=None,
        partner_code=None,
        product_code=None,
        period_start=None,
        period_end=None,
    ):
        """Select faostat data for the given arguments

        :param str table: name of the database table to select from
        :param list or str reporter: list of reporter names
        :param list or str partner: list of partner names
        :param list or str product: list of product names
        :param list or str element: list of element names
        :param list or int or str reporter_code: list of reporter codes
        :param list or int or str partner_code: list of partner codes
        :param list or int or str product_code: list of product codes
        :param int period_start: integer for filtering data from start year
        :param int period_end: integer for filtering data up to end year
        :return: A data frame of trade flows

        Note that the search for reporter and partner will be based on perfect
        matches whereas product can be partial matches.

        For example select crop production data for 2 countries

            >>> from biotrade.faostat import faostat
            >>> db = faostat.db
            >>> cp1 = db.select(table="crop_production",
            >>>                 reporter=["Portugal", "Estonia"])

        Select the same data using the country codes instead of the country names

            >>> cp2 = db.select(table="crop_production",
            >>>                 reporter_code=[63, 174])

        Compare values

            >>> index = ["reporter", "product", "element", "year"]
            >>> cp1.sort_values(index, inplace=True, ignore_index=True)
            >>> cp2.sort_values(index, inplace=True, ignore_index=True)
            >>> cp1.equals(cp2)

        Select forestry trade flows data reported by Austria with all partner countries:

            >>> ft_aut = faostat.db.select(table="forestry_trade",
            >>>                            reporter=["Austria"])

        Select forestry trade flows data reported by all countries, with
        Austria as a partner country:

            >>> ft_aut_p = faostat.db.select(table="forestry_trade",
            >>>                              partner=["Austria"])

        Select crop trade flows reported by the Netherlands where Brazil was a
        partner

            >>> ct_nel_bra = db.select(table="crop_trade",
            >>>                        reporter="Netherlands",
            >>>                        partner="Brazil")

        Select the mirror flows reported by Brazil, where the Netherlands was a partner

            >>> ct_bra_bel = db.select(table="crop_trade",
            >>>                        reporter="Brazil",
            >>>                        partner="Netherlands")

        Select crop production where products contain the word "soy"

            >>> from biotrade.faostat import faostat
            >>> db = faostat.db
            >>> soy_prod = db.select(table="crop_production",
            >>>                      product = "soy")

        Select soybeans and oil soybean production using their product codes

            >>> soy_prod_2 = db.select(table="crop_production",
            >>>                        product_code = [236, 237])

        Compare the two data frame. Sort them first because rows are in a
        different order.

            >>> index = ["reporter", "product", "element", "year"]
            >>> soy_prod_2.sort_values(index, inplace=True, ignore_index=True)
            >>> soy_prod.sort_values(index, inplace=True, ignore_index=True)
            >>> soy_prod.equals(soy_prod_2)

        Select crop trade where products contain the word "soy"

            >>> soy_trade = db.select(table="crop_trade", product = "soy")

        Select crop production where products contain the words in the list

            >>> products_of_interest = ["soy", "palm", "sun", "oil", "rapeseed"]
            >>> veg_oil = db.select(table="crop_production",
            >>>                      product = products_of_interest)
            >>> for e in veg_oil["element"].unique():
            >>>     print(e)
            >>>     print(veg_oil[veg_oil.element == e]["product"].unique())

        Select crop trade where products contain the words in the list

            >>> veg_oil = db.select(table="crop_trade",
            >>>                     product = products_of_interest)

        """
        table = self.tables[table]
        # Change character or integer arguments to lists suitable for a
        # column.in_() clause or for a list comprehension.
        if isinstance(reporter, str):
            reporter = [reporter]
        if isinstance(partner, str):
            partner = [partner]
        if isinstance(product, str):
            product = [product]
        if isinstance(element, str):
            element = [element]
        if isinstance(reporter_code, (int, str)):
            reporter_code = [reporter_code]
        if isinstance(partner_code, (int, str)):
            partner_code = [partner_code]
        if isinstance(product_code, (int, str)):
            product_code = [product_code]
        # Build the select statement
        stmt = table.select()
        if reporter is not None:
            stmt = stmt.where(table.c.reporter.in_(reporter))
        if partner is not None:
            stmt = stmt.where(table.c.partner.in_(partner))
        if product is not None:
            stmt = stmt.where(or_(table.c.product.ilike(f"%{p}%") for p in product))
        if element is not None:
            stmt = stmt.where(table.c.element.in_(element))
        if reporter_code is not None:
            stmt = stmt.where(table.c.reporter_code.in_(reporter_code))
        if partner_code is not None:
            stmt = stmt.where(table.c.partner_code.in_(partner_code))
        if product_code is not None:
            stmt = stmt.where(table.c.product_code.in_(product_code))
        if period_start is not None:
            stmt = stmt.where(table.c.period >= period_start)
        if period_end is not None:
            stmt = stmt.where(table.c.period <= period_end)
        # Query the database and return a data frame
        df = self.read_sql_query(stmt)
        return df

    def agg_reporter_partner_eu_row(
        self,
        table,
        product_code,
    ):
        """
        Aggregate EU27 and ROW both on the reporter and eventually partner side
        as regard faostat tables, using sqlalchemy statements for specific
        product code list

        :param (list) product_code, list of integer product codes
        :return (DataFrame) df, containing the aggregations for the product codes

        As example, query the crop trade table and aggregate EU27 and ROW data
        both reporter and partner side for product code equal to 236,
        corresponding to soybean commodity

            >>> from biotrade.faostat import faostat
            >>> db = faostat.db
            >>> df = db.agg_reporter_partner_eu_row("crop_trade", product_code=[236])

        """
        # Table to select of raw_faostat schema
        table = self.tables[table]
        # Select all columns where product code is specified by function argument product_code
        # If product code list is None return an error
        table_stmt = table.select()
        if product_code is None:
            raise ValueError("Specify product code list")
        else:
            table_stmt = table_stmt.where(table.c.product_code.in_(product_code))
        # Render the selection queryable again and rename it "table_selection"
        table_selection = table_stmt.subquery().alias("table_selection")
        # Select country table of raw_faostat_schema
        country_table = self.tables["country"]
        # Select two columns of country table and rename them for reporter selection
        reporter_stmt = select(
            [
                country_table.c.country_code.label("reporter_code"),
                country_table.c.eu27.label("reporter_eu27"),
            ]
        )
        # Render the selection queryable again and rename it "reporter_selection"
        reporter_selection = reporter_stmt.subquery().alias("reporter_selection")
        # Outer left join on reporter side
        join_stmt = table_selection.join(
            reporter_selection,
            table_selection.c.reporter_code == reporter_selection.c.reporter_code,
            isouter=True,
        )
        # If the selected table contains the partner column, the aggregation EU/ROW is performed also from this side
        if "partner" in table.c.keys():
            # Select two columns of country table and rename them for partner selection
            partner_stmt = select(
                [
                    country_table.c.country_code.label("partner_code"),
                    country_table.c.eu27.label("partner_eu27"),
                ]
            )
            # Render the selection queryable again and rename it "partner_selection"
            partner_selection = partner_stmt.subquery().alias("partner_selection")
            # Second outer left join on partner side
            join_stmt = join_stmt.join(
                partner_selection,
                table_selection.c.partner_code == partner_selection.c.partner_code,
                isouter=True,
            )
        # Select the join statement and rename it "join_selection"
        join_selection = select(join_stmt).alias("join_selection")
        # Columns to retain from the join selection
        join_columns = [
            join_selection.c.product_code,
            join_selection.c.product,
            join_selection.c.element,
            join_selection.c.unit,
            join_selection.c.year,
            join_selection.c.reporter_eu27,
            func.sum(join_selection.c.value).label("value"),
        ]
        # Column to retain from the join selection for partner side
        if "partner" in table.c.keys():
            join_columns.insert(-1, join_selection.c.partner_eu27)
        # Select column subset of join selection to return for dataframe
        stmt = select(join_columns)
        # Delete from column list the column not used for the groupby
        del join_columns[-1]
        # Aggregate by columns the join selection
        stmt = stmt.group_by(*join_columns)
        # Return the dataframe from the query to db
        df = self.read_sql_query(stmt)
        return df

    def extract_product_names_codes(self, table_list):
        """
        Extract product names and product codes from tables of
        Faostat db. Duplicates are dropped out.

        :param (table_list) tables, list of Faostat table names
        :return (DataFrame) faostat_products, containing product names and codes

        For example, obtain product codes and names from crop_trade and
        crop_production Faostat tables

        >>> from biotrade.faostat import faostat
        >>> table_list = ["crop_production", "crop_trade"]
        >>> faostat_products = faostat.db.extract_product_names_codes(table_list)

        """
        faostat_products = pandas.DataFrame()
        # Loop over tables
        for table in table_list:
            # Select faostat table
            faostat_table = self.tables[table]
            # Select product names and codes from Faostat table
            stmt = (
                faostat_table.select()
                .with_only_columns(
                    [faostat_table.c.product_code, faostat_table.c.product]
                )
                .distinct(faostat_table.c.product_code, faostat_table.c.product)
            )
            # Retrieve dataset
            table_products = self.read_sql_query(stmt)
            faostat_products = pandas.concat(
                [faostat_products, table_products], ignore_index=True
            )
        # Drop duplicates
        faostat_products.drop_duplicates(
            subset="product_code", ignore_index=True, inplace=True
        )
        return faostat_products


class DatabaseFaostatPostgresql(DatabaseFaostat):
    """Database using the PostgreSQL engine use the same database for all data sources
    a separate schema for each data source

    The connection URL is defined in an environment variable, for example by
    adding the following to  your .bash_aliases or .bash_rc:

        export BIOTRADE_DATABASE_URL="postgresql://rdb@localhost/biotrade"
    """

    database_url = database_url
    schema = "raw_faostat"


class DatabaseFaostatSqlite(DatabaseFaostat):
    """Database using the SQLite engine"""

    database_url = f"sqlite:///{data_dir}/faostat/faostat.db"
    schema = "main"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Paul Rougieux and Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

JRC biomass Project.
Unit D1 Bioeconomy.

"""
import pandas

CONVERSION_FACTORS_LEVEL1 = pandas.DataFrame(
    {
        "product": [
            "industrial_roundwood",
            "wood_fuel",
            "sawnwood",
            "wood_based_panels",
            "paper_and_paperboard",
        ],
        "coef": [1, 1, 1.88, 1.5, 3.5],
        "unit_converted": [
            "m3 roundwood/m3",
            "m3 roundwood/m3",
            "m3 roundwood/m3",
            "m3 roundwood/m3",
            "m3 roundwood/t",
        ],
    }
)


def convert_to_eq_rwd_level_1(
    df, conv_factors=CONVERSION_FACTORS_LEVEL1, selected_units=None
):
    """Convert level one products to their volume equivalent roundwood
    All other products are ignored.

    Usage:

    >>> from biotrade.faostat.convert import convert_to_eq_rwd_level_1
    >>> from biotrade.faostat import faostat
    >>> fp_ita = faostat.db.select(table="forestry_production",
    >>>                                   reporter=["Italy"])
    >>> fp_ita_eqr = convert_to_eq_rwd_level_1(fp_ita)

    """
    if selected_units is None:
        selected_units = ["m3", "tonnes"]
    selected_products = conv_factors["product"].to_list()
    # Keep only products present in the conv_factors table
    # Compute their volume equivalent roundwood
    dfeqr = (
        df.query("product in @selected_products and unit in @selected_units")
        .merge(conv_factors, on="product")
        .assign(value_eqrwd=lambda x: x["value"] * x["coef"])
    )
    return dfeqr



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Paul Rougieux and Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

JRC biomass Project.
Unit D1 Bioeconomy.

Usage update the World Bank data :

    >>> from biotrade.world_bank import world_bank
    >>> world_bank.pump.update(["indicator", "indicator_name"])

"""
from pathlib import Path
from zipfile import ZipFile
import tempfile
import logging

try:
    import requests
except Exception as e:
    msg = (
        "Failed to import requests, you will not be able to load data from World Bank,"
    )
    msg += "but you can still use other methods.\n"
    print(msg, str(e))
import shutil
import pandas as pd
import csv

# Internal modules
from biotrade.common.url_request_header import HEADER
from biotrade.common.sanitize import sanitize_variable_names


class Pump:
    """
    Download World Bank data and store it locally in a database

    The pump can perform the following tasks:
        1. Download compressed csv files from the World Bank.
        2. Read the compressed csv files into pandas data frames.
        3. Transfer the data frames to a database.

    An update performs all 3 tasks, it download files, reads them in chunks and
    stores them in the database

        >>> from biotrade.world_bank import world_bank
        >>> world_bank.pump.update()

    Each task individually

       >>> world_bank.pump.download_zip_csv()

    """

    # Log debug and error messages
    logger = logging.getLogger("biotrade.world_bank")
    # Define URL request headers
    header = HEADER
    # URL to load data from the website
    url_bulk = "https://databank.worldbank.org/data/download/WDI_CSV.zip?_gl=1*d0bq8i*_gcl_au*MTQ3NTE1NTk2OC4xNzIxMjkwNDE1"
    # Destination zip and csv file names
    zip_file_name = "WDI_CSV.zip"
    datasets = {
        "indicator": "WDICSV.csv",
        "indicator_name": "WDICSV.csv",
    }

    def __init__(self, parent):
        # Default attributes #
        self.parent = parent
        self.db = self.parent.db
        self.data_dir = self.parent.data_dir
        # Number of lines to read from csv files before transform years to long format
        self.chunk_size = 10**4

    def download_zip_csv(self):
        """download a compressed csv file containing all World Bank indicators"""
        # Load the zip file to the data directory
        output_file = self.data_dir / self.zip_file_name
        self.logger.info("Downloading data from:\n %s", self.url_bulk)
        response = requests.get(url=self.url_bulk, headers=self.header, stream=True)
        with open(output_file, "wb") as out_file:
            print(f"HTTP response code: {response.status_code}")
            shutil.copyfileobj(response.raw, out_file)

    def transfer_csv_to_db_in_chunks(self, short_name, chunk_size, reformatting):
        """Read the World Bank zip csv file and transfer large long format CSV
        file to the database in chunks so that a data frame with millions of
        rows doesn't overload the memory."""
        temp_dir = Path(tempfile.TemporaryDirectory().name)
        try:
            # Unzip the CSV and write it to a temporary file on disk
            zip_file = ZipFile(self.data_dir / self.zip_file_name)
            zip_file.extractall(temp_dir)
            # Obtain the name of csv file to read
            csv_file = temp_dir / self.datasets[short_name]
            # Name of csv file columns to use
            id_columns = [
                "Country Name",
                "Country Code",
                "Indicator Name",
                "Indicator Code",
            ]
            if reformatting:
                # Retrieve series file to append unit column for indicator table
                unit_file = temp_dir / "WDISeries.csv"
                df_unit = pd.read_csv(unit_file)
                # Rename column same as csv file
                df_unit.rename(columns={"Series Code": "Indicator Code"}, inplace=True)
                # Keep only columns needed for the merge
                df_unit = df_unit[["Indicator Code", "Unit of measure"]]
            else:
                # Do not split the dataframe into chunks (len(df) = 383572)
                chunk_size = 10**6
            # Test if the file is corrupted
            with open(csv_file, "r") as csvfile:
                # Detect the delimiter
                dialect = csv.Sniffer().sniff(csvfile.read(1024))
                # Place the reader at the beginning
                csvfile.seek(0)
                # Read the file
                reader = csv.reader(csvfile, dialect)
                header = next(reader)
                for row in reader:
                    pass
        # Zip file corrupted
        except Exception as e:
            self.db.logger.warning(
                f"File for {short_name} table is not available due to {e}.\n Unable to pump {short_name} data."
            )
            if temp_dir.exists():
                # Remove temporary directory
                shutil.rmtree(temp_dir)
            return
        # Drop and recreate the table
        table = self.db.tables[short_name]
        table.drop(self.db.engine)
        self.db.create_if_not_existing(table)
        # Read the csv file, transform the dataframe and upload data to the database
        for df_chunk in pd.read_csv(csv_file, chunksize=chunk_size):
            # Remove unnamed columns
            df_chunk.drop(df_chunk.filter(regex="Unnamed"), axis=1, inplace=True)
            if reformatting:
                # Reformatting year columns into long format
                df_chunk = df_chunk.melt(
                    id_vars=id_columns, var_name="year", value_name="value"
                )
                # Merge with unit column
                df_chunk = df_chunk.merge(
                    df_unit,
                    on=["Indicator Code"],
                    how="left",
                )
            else:
                # Get unique indicator names and codes
                df_chunk = df_chunk.drop_duplicates(
                    subset=["Indicator Name", "Indicator Code"]
                )
            # Rename columns and keep only those needed for the db table
            df_chunk = sanitize_variable_names(
                df_chunk, "world_bank", self.db, short_name
            )
            print(df_chunk.head(1))
            # Append chunk to the db
            self.db.append(df=df_chunk, table=short_name)
        if temp_dir.exists():
            # Remove temporary directory
            shutil.rmtree(temp_dir)

    def transfer_to_db(self, datasets, skip_confirmation=False):
        """Transfer from a csv file to the database by replacing the table
        content with the content of the zipped CSV files. Database field types
        are determined in world_bank.db.

        :param list datasets: list of dataset names, whose keys should be in
            the world_bank.pump.datasets and world_bank.db.tables dictionaries
        :param boolean skip_confirmation: ask confirmation as default (False)
            before deleting tables
        :return: Nothing

        Update all world bank indicators from the bulk file into the "indicator" table:

            >>> from biotrade.world_bank import world_bank
            >>> world_bank.pump.transfer_to_db("indicator")

        Use a larger chunk size

            >>> world_bank.pump.chunk_size = 10 ** 5

        """
        # Make datasets a list
        if isinstance(datasets, str):
            datasets = [datasets]
        if not skip_confirmation:
            if not self.confirm_db_table_deletion(datasets):
                return
        for table_name in datasets:
            # Keep as default reformatting csv to long years format
            reformatting = True
            if table_name == "indicator_name":
                # Do not reformat the csv file
                reformatting = False
            # Transfer the compressed CSV file to the database
            self.transfer_csv_to_db_in_chunks(table_name, self.chunk_size, reformatting)

    def update(self, datasets, skip_confirmation=False):
        """Update the given datasets by downloading them from World Bank Data and
        transferring them to the database

        :param list or str datasets: list of dataset names, whose keys should
            be in the world_bank.pump.datasets and world_bank.db.tables dictionaries

        Usage:

            >>> from biotrade.world_bank import world_bank
            >>> world_bank.pump.update(["indicator"])

        """
        # Make datasets a list
        if isinstance(datasets, str):
            datasets = [datasets]
        # Confirmation message
        if not skip_confirmation:
            if not self.db.confirm_db_table_deletion(datasets):
                return
        # Download compressed zip bulk dataset from World Bank Data to data folder
        self.download_zip_csv()
        # Transfer to the database
        # Skip confirmation because we already confirmed above
        self.transfer_to_db(datasets, skip_confirmation=True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Paul Rougieux and Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

JRC biomass Project.
Unit D1 Bioeconomy.

You can use this object at the ipython console with the following examples.
Load population data https://data.worldbank.org/indicator/SP.POP.TOTL

    >>> from biotrade.world_bank import world_bank
    >>> pop = world_bank.db.select("indicator", indicator_code="SP.POP.TOTL")

Select available indicator time series for Italy and Germany and display those
that contain the word "gdp".

    >>> from biotrade.world_bank import world_bank
    >>> indicators = world_bank.db.select("indicator", reporter=["Germany", "Italy"])
    >>> icode = indicators["indicator_code"]
    >>> icode[icode.str.contains("gdp", case=False)].unique()

Select the GDP in constant USD of 2015 and display the country groups and
countries with the highest GDP
https://data.worldbank.org/indicator/NY.GDP.MKTP.KD

    >>> gdpk2015 = world_bank.db.select("indicator", indicator_code="NY.GDP.MKTP.KD")
    >>> gdpk2015.groupby('reporter')["value"].max().sort_values(ascending=False).head(20)

"""
# First party modules
import logging

# Third party modules
import pandas
from sqlalchemy import Integer, Float, Text, UniqueConstraint
from sqlalchemy import Table, Column
from sqlalchemy import MetaData
from sqlalchemy import create_engine, inspect

# from sqlalchemy.sql import func
from sqlalchemy.schema import CreateSchema
from sqlalchemy_utils import database_exists, create_database

# Internal modules
from biotrade import data_dir, database_url
from biotrade.common.database import Database


class DatabaseWorldBank(Database):
    """
    Database to store World Bank Indicators data

    Database update methods are in the world_bank.pump object.

    """

    # To be overwritten by the children
    database_url = None
    schema = None

    # Log debug and error messages
    logger = logging.getLogger("biotrade.comtrade")

    def __init__(self, parent):
        # Default attributes #
        self.parent = parent
        # Database configuration
        self.engine = create_engine(self.database_url)
        # Create the database if it doesn't exist
        if not database_exists(self.engine.url):
            create_database(self.engine.url)

        # SQL Alchemy metadata
        self.metadata = MetaData(schema=self.schema)
        self.metadata.bind = self.engine
        self.inspector = inspect(self.engine)

        # Create the schema if it doesn't exist.
        # Exclude the SQLite engine because there is only a default schemas for that engine.
        # And the SQLite dialect doesn't have a has_schema() method.
        if hasattr(self.engine.dialect, "has_schema") and callable(
            getattr(self.engine.dialect, "has_schema")
        ):
            with self.engine.connect() as conn:
                if not self.engine.dialect.has_schema(conn, self.schema):
                    conn.execute(CreateSchema(self.schema))
                    conn.commit()

        # Describe table metadata
        self.indicator = self.describe_indicator_table(name="indicator")
        self.indicator_name = self.describe_indicator_name_table(name="indicator_name")
        self.tables = {
            "indicator": self.indicator,
            "indicator_name": self.indicator_name,
        }
        # Create tables if they don't exist
        for table in self.tables.values():
            self.create_if_not_existing(table)

    def describe_indicator_table(self, name):
        """Define the metadata of a table containing indicator data."""
        table = Table(
            name,
            self.metadata,
            Column("reporter_code", Text, index=True),
            Column("reporter", Text),
            Column("indicator_code", Text, index=True),
            Column("year", Integer),
            Column("unit", Text),
            Column("value", Float),
            # Unit not in the unique constraint due to some nan values
            UniqueConstraint(
                "reporter_code",
                "indicator_code",
                "year",
            ),
            schema=self.schema,
        )
        return table

    def describe_indicator_name_table(self, name):
        """Define the metadata of a table containing indicator name data."""
        table = Table(
            name,
            self.metadata,
            Column("indicator_code", Text, index=True),
            Column("indicator_name", Text),
            UniqueConstraint(
                "indicator_code",
            ),
            schema=self.schema,
        )
        return table

    def select(
        self,
        table,
        reporter=None,
        reporter_code=None,
        indicator_code=None,
        year_start=None,
        year_stop=None,
    ):
        """
        Select world bank data for the given arguments

        :param str table: name of the database table to select from
        :param list or str reporter: list of reporter names
        :param list or int reporter_code: list of reporter codes
        :param list or int indicator_code: list of indicator codes
        :param int year_start: year from which data are retrieved
        :param int year_stop: year end data
        :return: A data frame of world bank indicators

        For example select population data https://data.worldbank.org/indicator/SP.POP.TOTL

            >>> from biotrade.world_bank import world_bank
            >>> pop = world_bank.db.select("indicator", indicator_code="SP.POP.TOTL")

        """
        table = self.tables[table]
        # Change string arguments to lists suitable for a
        # column.in_() clause
        if isinstance(reporter, str):
            reporter = [reporter]
        if isinstance(reporter_code, str):
            reporter_code = [reporter_code]
        if isinstance(indicator_code, str):
            indicator_code = [indicator_code]
        # Build the select statement
        stmt = table.select()
        if reporter is not None:
            stmt = stmt.where(table.c.reporter.in_(reporter))
        if reporter_code is not None:
            stmt = stmt.where(table.c.reporter_code.in_(reporter_code))
        if indicator_code is not None:
            stmt = stmt.where(table.c.indicator_code.in_(indicator_code))
        if year_start is not None:
            stmt = stmt.where(table.c.year >= year_start)
        if year_stop is not None:
            stmt = stmt.where(table.c.year <= year_stop)
        # Query the database and return a data frame
        with self.engine.connect() as conn:
            df = pandas.read_sql_query(stmt, conn)
        return df


class DatabaseWorldBankPostgresql(DatabaseWorldBank):
    """Database using the PostgreSQL engine use the same database for all data sources
    a separate schema for each data source

    The connection URL is defined in an environment variable, for example by
    adding the following to  your .bash_aliases or .bash_rc:

        export BIOTRADE_DATABASE_URL="postgresql://rdb@localhost/biotrade"
    """

    database_url = database_url
    schema = "raw_world_bank"


class DatabaseWorldBankSqlite(DatabaseWorldBank):
    """Database using the SQLite engine"""

    database_url = f"sqlite:///{data_dir}/world_bank/world_bank.db"
    schema = "main"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Paul Rougieux and Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

JRC Biomass Project.
Unit D1 Bioeconomy.
"""

# Third party modules
import pandas
import numpy as np

# Internal modules
from biotrade import module_dir
from biotrade.faostat.aggregate import agg_trade_eu_row

PRODUCTS = pandas.read_csv(
    module_dir / "config_data/faostat_forestry_production_short_names.csv"
)
SELECTED_PRODUCTS = [
    "ir_c",
    "ir_nc",
    "pp",
    "sw",
    "sw_c",
    "sw_nc",
    "wp",
    "fw_c",
    "fw_nc",
    "wbp",
    "wpp",
]
PRODUCTS = PRODUCTS[PRODUCTS.product_short.isin(SELECTED_PRODUCTS)].copy()

ELEMENTS = pandas.DataFrame(
    {
        "element": [
            "production",
            "import_quantity",
            "import_value",
            "export_quantity",
            "export_value",
        ],
        "element_short": ["prod", "imp", "imp_usd", "exp", "exp_usd"],
    }
)


class HwpCountry:
    """
    This object gives access to the Harvested Wood Products data for one country.

        >>> from biotrade.common.country import Country
        >>> aut = Country("Austria")
        >>> aut.hwp.production

    """

    def __repr__(self):
        return '%s object name "%s"' % (self.__class__, self.country_name)

    def __init__(self, parent):
        """Get the country name from this object's parent class."""
        self.country_name = parent.country_name
        self.faostat_country = parent.faostat
        self.products = PRODUCTS
        self.selected_products = SELECTED_PRODUCTS
        self.elements = ELEMENTS

    @property
    def production(self):
        """FAOSTAT forestry production data for the selected products

        >>> from biotrade.common.country import Country
        >>> aut = Country("Austria")
        >>> aut.hwp.production

        Reshape to wide format using short column names

        >>> prod_wide = (aut.hwp.production
        >>>              .pivot(index=["reporter", "year"],
        >>>                     columns=["prod_elem"],
        >>>                     values="value")
        >>>             )

        """
        df = self.faostat_country.forestry_production.copy()
        # Prepare shorter column names combination of product and element
        df = df.merge(PRODUCTS, on=["product", "product_code"], how="inner")
        df = df.merge(ELEMENTS, on=["element"], how="inner")
        df = df[
            df.element.isin(["production", "import_quantity", "export_quantity"])
        ].copy()
        df["prod_elem"] = df["product_short"] + "_" + df["element_short"]
        df = df.drop(columns=["element_short"])
        return df

    @property
    def trade(self):
        """Bilateral forestry trade data from FAOSTAT

        >>> from biotrade.common.country import Country
        >>> aut = Country("Austria")
        >>> aut.hwp.trade

        """
        df = self.faostat_country.forestry_trade.copy()
        return df

    @property
    def trade_eu_row(self, index_side="partner"):
        """FAOSTAT forestry trade data aggregated for EU and Non EU partners

        >>> from biotrade.common.country import Country
        >>> aut = Country("Austria")
        >>> aut.hwp.trade_eu_row

        >>> from biotrade.common.country import Country
        >>> bel = Country("Belgium")
        >>> bel.hwp.trade_eu_row

        """
        df = agg_trade_eu_row(self.trade, grouping_side=index_side)
        df = df.merge(PRODUCTS, on=["product", "product_code"], how="inner")
        return df

    @property
    def production_estimated_century(self):
        """Production estimated from 1901 to FAOSTAT start year"""
        df = self.production
        base_year = df["year"].min()
        df_base = df[df["year"] == base_year]
        df_past = pandas.concat([df_base] * 60, ignore_index=True)
        # Generate a list of number for each product group
        df_past["n"] = 1
        df_past["n"] = df_past.groupby(["prod_elem"])["n"].cumsum()
        # Use the IPCC estimation approach
        df_past = df_past.rename(columns={"value": "base_value"})
        df_past["value"] = df_past["base_value"] * np.exp(0.0151 * -df_past["n"])
        df_past = df_past.sort_values(["product", "year"])
        # TODO: finish the implementation of this method by correcting years



""" Aggregate Comtrade at the 4 digit level


For example aggregate comtrade data from the 6 to the 4 digit level and store
the resulting aggregated data frame to a parquet file.

    >>> import biotrade
    >>> from biotrade.comtrade.aggregate import agg_4d
    >>> # Create a directory to save parquet files
    >>> comtrade_4d_dir = biotrade.data_dir / "comtrade" / "4d"
    >>> comtrade_4d_dir.mkdir(exist_ok=True)
    >>> # Load data in a loop, aggregate and save to parquet files
    >>> for this_code in ["44", "47", "48", "49"]:
    >>>     print(f"Preparing {this_code}")
    >>>     df = agg_4d(product_code_start=this_code)
    >>>     df.to_parquet(comtrade_4d_dir / f"ft{this_code}.parquet")

df.info() shows a memory usage of 733 MB
df_agg.info() shows a memory usage of 57 MB
The ratio of number of lines is only half: len(df_agg) / len(df) = 0.57.
The large difference in size is because we dropped many columns in the aggregation.

"""

from biotrade.comtrade import comtrade


def agg_4d(product_code_start, table=None):
    """ "Aggregate the input data frame to 4 digit level for all products starting with the given code.
    The table argument defaults to "yearly".
    """
    if table is None:
        table = "yearly"

    # Load Global trade data in a loop
    df = comtrade.db.select(
        table, product_code_start=product_code_start, period_start=2010
    )

    df["product_code_4d"] = df["product_code"].str[:4]
    index = [
        "reporter_code",
        "reporter",
        "partner_code",
        "partner",
        "product_code_4d",
        "period",
        "year",
        "unit",
        "flow_code",
        "flow",
    ]
    df_agg = df.groupby(index)[["net_weight", "quantity", "trade_value"]].agg("sum")

    del df
    return df_agg



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Paul Rougieux and Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

JRC biomass Project.
Unit D1 Bioeconomy.


"""

import numpy as np
import re


def sanitize_variable_names(df, column_renaming, db, short_name):
    """Sanitize column names using a mapping table."""
    # Rename columns to snake case
    df.rename(columns=lambda x: re.sub(r"\W+", "_", str(x)).lower(), inplace=True)
    # Columns of the db table
    db_table_cols = db.tables[short_name].columns.keys()
    # Original column names
    cols_to_check = db.parent.column_names[
        db.parent.column_names["biotrade"].isin(db_table_cols)
    ][column_renaming].tolist()
    # Check columns which have changed in the input source
    cols_to_change = set(cols_to_check).difference(df.columns)
    # If column names have changed raise an error
    if cols_to_change:
        raise ValueError(
            f"The following columns \n{list(cols_to_change)}\nhave changed in the input source {column_renaming}.\nUpdate config_data/column_names.csv before updating table {short_name}"
        )
    # Map columns using the naming convention defined in self.column_names
    mapping = db.parent.column_names.set_index(column_renaming).to_dict()["biotrade"]
    # Discard nan keys of mapping dictionary
    mapping.pop(np.nan, None)
    # Filter df selecting only columns for db
    df = df[cols_to_check].copy()
    # Rename columns using the naming convention defined in self.column_names
    df.rename(columns=mapping, inplace=True)
    return df



# Importing the CronTab class from the module
from crontab import CronTab
from dotenv import dotenv_values
# Using patasel user
cron = CronTab(user="patasel")
# Delete previous jobs and environment variables
cron.remove_all()
cron.env.clear()
cron.write()
# Import environment variables and assign them to the crontab
envs = dotenv_values("/storage/SUSBIOM-TRADE/env_var/.env")
for (name, value) in envs.items():
    cron.env[name] = value
# Update faostat data at 03:45 on day-of-month 8 in every 3rd month
job = cron.new(command="""/storage/SUSBIOM-TRADE/conda/susbiom_trade_env/bin/python -c "from biotrade.faostat import faostat;faostat.pump.update(['country', 'crop_production', 'crop_trade', 'food_balance', 'forestry_production', 'forestry_trade', 'forest_land', 'land_use', 'land_cover'],skip_confirmation=True)" >> /eos/jeodpp/data/projects/SUSBIOM-TRADE/biotrade_data/crontab_faostat.log 2>&1""")
job.minute.on(45)
job.hour.on(3)
job.day.on(8)
job.month.every(3)
cron.write()
# Update comtrade yearly data at 04:45 on day-of-month 8 in every 3rd month
job = cron.new(command="""/storage/SUSBIOM-TRADE/conda/susbiom_trade_env/bin/python -c "from biotrade.comtrade import comtrade;comtrade.pump.update_db(table_name = 'yearly', frequency = 'A')" >> /eos/jeodpp/data/projects/SUSBIOM-TRADE/biotrade_data/crontab_comtrade_yearly.log 2>&1""")
job.minute.on(45)
job.hour.on(4)
job.day.on(8)
job.month.every(3)
cron.write()
# Update comtrade monthly data at 04:45 on day-of-month 2 and 16
job = cron.new(command="""/storage/SUSBIOM-TRADE/conda/susbiom_trade_env/bin/python -c "from biotrade.comtrade import comtrade;comtrade.pump.update_db(table_name = 'monthly', frequency = 'M')" >> /eos/jeodpp/data/projects/SUSBIOM-TRADE/biotrade_data/crontab_comtrade_monthly.log 2>&1""")
job.minute.on(45)
job.hour.on(4)
job.day.on(2,16)
cron.write()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Paul Rougieux and Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

JRC biomass Project.
Unit D1 Bioeconomy.

You can use this object at the ipython console with the following examples.

Get yearly Comtrade data at the 2 digit level.
Load the complete table into a pandas data frame.

    >>> import pandas
    >>> from biotrade.comtrade import comtrade
    >>> db = comtrade.db_pgsql
    >>> df = pandas.read_sql_table("yearly_hs2", db.engine, schema="raw_comtrade")

Select monthly oak trade data using the select method defined here.

    >>> from biotrade.comtrade import comtrade
    >>> oak = comtrade.db.select("monthly", product_code="440791")

Select data for the year 2017 using an SQL Alchemy select statement (lower level). Return results
using an SQL Alchemy cursor or with a pandas data frame:

    >>> year = db.yearly_hs2.columns.get("year")
    >>> statement = db.yearly_hs2.select().where(year == 2017)
    >>> # Get results as a list from the cursor
    >>> with db.engine.connect() as connection:
    >>>     result = connection.execute(statement)
    >>> for row in result:
    >>>     print(row)
    >>> # Legacy cursor
    >>> result_2 = db.engine.execute(statement).fetchall()
    >>> # Results as a data frame
    >>> df_2017 = pandas.read_sql_query(statement, db.engine)

Note the main variables of interest are "net_weight" and "trade_value". A query
of the monthly data shows that "quantity" is always empty:

    select * from raw_comtrade.monthly where quantity is not null limit 4;
    # Returns 0 rows

The database structure is defined as SQLAlchemy objects, then later translated
in either SQLite or PostGreSQL dialect. Show the database structure as defined
in PostGreSQL:

    pg_dump --schema-only -t raw_comtrade.yearly $BIOTRADE_DATABASE_URL > table_structure_dump.sql

"""
# First party modules
import logging
import pandas
import re

# Third party modules
from sqlalchemy import Integer, Float, Text, UniqueConstraint, DateTime, Boolean
from sqlalchemy import Table, Column, MetaData, and_, or_
from sqlalchemy import create_engine, inspect, select
from sqlalchemy.schema import CreateSchema

# Internal modules
from biotrade import data_dir, database_url
from biotrade.common.database import Database


class DatabaseComtrade(Database):
    """
    Database to store UN Comtrade data.
    """

    # To be overwritten by the children
    database_url = None
    schema = None

    # Log debug and error messages
    logger = logging.getLogger("biotrade.comtrade")

    def __init__(self, parent):
        # Default attributes #
        self.parent = parent
        # Database configuration
        self.engine = create_engine(self.database_url)
        # SQL Alchemy metadata
        self.metadata = MetaData(schema=self.schema)
        self.metadata.bind = self.engine
        self.inspector = inspect(self.engine)

        # Create the schema if it doesn't exist.
        # Exclude the SQLite engine because there is only a default schemas for that engine.
        # And the SQLite dialect doesn't have a has_schema() method.
        if hasattr(self.engine.dialect, "has_schema") and callable(
            getattr(self.engine.dialect, "has_schema")
        ):
            with self.engine.connect() as conn:
                if not self.engine.dialect.has_schema(conn, self.schema):
                    conn.execute(CreateSchema(self.schema))
                    conn.commit()

        # Describe table metadata and create them if they don't exist
        # Product table
        self.product = self.describe_product_table(name="product")
        # Reporter/Partner table
        self.reporter = self.describe_country_table(name="reporter")
        self.partner = self.describe_country_table(name="partner")
        # Additional info tables
        self.quantity = self.describe_additional_info_table(name="quantity_unit")
        self.flow = self.describe_additional_info_table(name="flow")
        self.transport = self.describe_additional_info_table(
            name="modality_of_transport"
        )
        self.custom = self.describe_additional_info_table(name="custom")
        # Trade tables with data at the HS 6 digit level
        self.monthly = self.describe_trade_table(name="monthly")
        self.yearly = self.describe_trade_table(name="yearly")
        # Trade table with data at the HS 2 digit level
        self.yearly_hs2 = self.describe_trade_table(name="yearly_hs2")
        self.tables = {
            "product": self.product,
            "reporter": self.reporter,
            "partner": self.partner,
            "quantity_unit": self.quantity,
            "flow": self.flow,
            "modality_of_transport": self.transport,
            "custom": self.custom,
            "monthly": self.monthly,
            "yearly": self.yearly,
            "yearly_hs2": self.yearly_hs2,
        }
        # Create tables if they don't exist
        for table in self.tables.values():
            self.create_if_not_existing(table)

    def describe_country_table(self, name):
        """Define the metadata of a table containing reporter/partner codes

        :param str table: name of the table to create
        """
        # Additional columns depending on the specific table
        table_dict = {
            "reporter": [
                Column("reporter_code", Integer),
                Column("reporter", Text),
                Column("reporter_note", Text),
                Column("reporter_iso2", Text),
                Column("reporter_iso", Text),
                UniqueConstraint(
                    "reporter_code",
                ),
            ],
            "partner": [
                Column("partner_code", Integer),
                Column("partner", Text),
                Column("partner_note", Text),
                Column("partner_iso2", Text),
                Column("partner_iso", Text),
                UniqueConstraint(
                    "partner_code",
                ),
            ],
        }
        # Common structure of the table
        table = Table(
            name,
            self.metadata,
            Column("id", Integer),
            Column("text", Text),
            Column("entry_effective_date", DateTime),
            Column("is_group", Boolean),
            Column("entry_expired_date", DateTime),
            schema=self.schema,
        )
        for col in table_dict[name]:
            table.append_column(col)
        return table

    def describe_product_table(self, name):
        """Define the metadata of a table containing product codes

        :param str table: name of the table to create
        """
        table = Table(
            name,
            self.metadata,
            Column("product_code", Text),
            Column("product_description", Text),
            Column("parent", Text),
            Column("is_leaf", Integer),
            Column("aggregate_level", Integer),
            UniqueConstraint(
                "product_code",
            ),
            schema=self.schema,
        )
        return table

    def describe_additional_info_table(self, name):
        """Define the metadata of tables containing additional info

        :param str table: name of the table to create
        """
        # Additional columns depending on the specific table
        table_dict = {
            "quantity_unit": [
                Column("unit_code", Integer),
                Column("unit_abbreviation", Text),
                Column("unit", Text),
                UniqueConstraint(
                    "unit_code",
                ),
            ],
            "flow": [
                Column("flow_code", Text),
                Column("flow", Text),
                UniqueConstraint(
                    "flow_code",
                ),
            ],
            "modality_of_transport": [
                Column("mode_of_transport_code", Integer),
                Column("mode_of_transport", Text),
                UniqueConstraint(
                    "mode_of_transport_code",
                ),
            ],
            "custom": [
                Column("custom_proc_code", Text),
                Column("custom_proc", Text),
                UniqueConstraint(
                    "custom_proc_code",
                ),
            ],
        }
        # Common structure of the table
        table = Table(
            name,
            self.metadata,
            schema=self.schema,
        )
        for col in table_dict[name]:
            table.append_column(col)
        return table

    def describe_trade_table(self, name):
        """Define the metadata of a table containing Comtrade data.

        :param str table: name of the table to create

        The unique constraint is a very important part of the table structure.
        It makes sure that there will be no duplicated flows.

        Alternatively a table metadata structure could be automatically loaded with:

            Table('yearly_hs2', self.metadata, autoload_with=self.engine)

        The python code below was originally generated with:

            sqlacodegen --schema raw_comtrade --tables yearly_hs2 postgresql://rdb@localhost/biotrade

        Note the "product" column is left empty, removed from the data frame
        before it is stored in the database because it would be too large. The
        text description of a product (commodity) is available in the product table.
        """
        table = Table(
            name,
            self.metadata,
            Column("dataset_code", Text),
            Column("classification_search_code", Text),
            Column("classification", Text),
            Column("is_original_classification", Integer),
            Column("frequency", Text),
            Column("ref_date", Integer),
            Column("year", Integer),
            Column("month", Integer),
            Column("period", Integer),
            Column("is_reported", Integer),
            Column("is_aggregated", Integer),
            Column("flow_code", Text),
            Column("reporter_code", Integer, index=True),
            Column("partner_code", Integer, index=True),
            Column("partner_2_code", Integer),
            Column("custom_proc_code", Text),
            Column("mode_of_supply_code", Text),
            Column("mode_of_transport_code", Integer),
            Column("product_code", Text, index=True),
            Column("product_type", Text),
            Column("unit_code", Integer),
            Column("quantity", Float),
            Column("is_quantity_estimated", Integer),
            Column("alt_qty_unit_code", Integer),
            Column("alt_qty", Float),
            Column("is_alt_quantity_estimated", Integer),
            Column("net_weight", Float),
            Column("is_net_weight_estimated", Integer),
            Column("gross_weight", Float),
            Column("is_gross_weight_estimated", Integer),
            Column("trade_value", Float),
            Column("cif_value", Float),
            Column("fob_value", Float),
            Column("flag", Integer),
            UniqueConstraint(
                "period",
                "flow_code",
                "reporter_code",
                "partner_code",
                "product_code",
                "flag",
            ),
            schema=self.schema,
        )
        return table

    def read_sql_query(self, stmt):
        """A wrapper around pandas.read_sql_query"""
        with self.engine.connect() as conn:
            df = pandas.read_sql_query(stmt, conn)
        return df

    def check_data_presence(
        self,
        table,
        start_year,
        end_year,
        frequency,
    ):
        """
        Query db table to check if data are present or not for a certain
        time period.

        :param (string) table, name of table to check data
        :param (int) start_year, initial period of the query
        :param (int) end_year, final year of the query
        :param (frequency) frequency, frequency of data
        :return (bool) check_presence, True if data are already inside
            db otherwise False

        For example check if monthly data from 2016 to 2022 are present:
            >>> from biotrade.comtrade import comtrade
            >>> data_check = comtrade.db.check_data_presence(
                    table = "monthly"
                    start_year = 2016,
                    end_year = 2022,
                    frequency = "M",
                )
        """
        check_presence = False
        # Select table
        table = self.tables[table]
        # Initial statement
        stmt = table.select()
        # Reformulate period as monthly or yearly frequency
        if frequency == "M":
            # Start from January of the start year
            start_period = int(str(start_year) + "01")
            # End in December of the end year
            end_period = int(str(end_year) + "12")
        elif frequency == "A":
            start_period = start_year
            end_period = end_year
        # Select data inside time from start period to end period
        if start_period is not None:
            stmt = stmt.where(table.c.period >= start_period)
        if end_period is not None:
            stmt = stmt.where(table.c.period <= end_period)
        # Read data from db to data frame and select the first one
        df = self.read_sql_query(stmt.limit(1))
        # If length of df is not zero, it means that data are inside db
        if len(df):
            check_presence = True
        return check_presence

    def delete_data(
        self,
        table,
        start_period,
        end_period,
    ):
        """
        Database method to delete rows from Comtrade db.

        :param (str) table, name of db table
        :param(int) start_period, from which time to delete rows
        :param(int) end_period, until which time to delete rows

        For example delete data from table "monthly" for January 2016:

            >>> from biotrade.comtrade import comtrade
            >>> comtrade.db.delete_data(
                    table = "monthly"
                    start_period = 201601,
                    end_period = 201601,
                )

        """
        # Table of db
        table = self.tables[table]
        assert start_period is not None
        assert end_period is not None
        # Construct delete statement
        stmt = table.delete().where(
            and_(
                table.c.period >= start_period,
                table.c.period <= end_period,
            )
        )
        # Execute delete statement
        with self.engine.connect() as conn:
            conn.execute(stmt)
        self.logger.info(
            "Delete data from database table %s, from %s to %s",
            table,
            start_period,
            end_period,
        )

    def select(
        self,
        table,
        reporter=None,
        partner=None,
        # TODO: use Comtrade.products.hs to enable a query on a product partial mapping
        reporter_code=None,
        partner_code=None,
        product_code=None,
        product_code_start=None,
        flow=None,
        period_start=None,
        period_end=None,
    ):
        """
        Select comtrade trade data for the given arguments, performing a join with additional information tables

        :param str table: name of the database table to select from
        :param list or str reporter: list of reporter names
        :param list or str partner: list of partner names
        :param list or int reporter_code: list of reporter codes
        :param list or int partner_code: list of partner codes
        :param list or int product_code: list of product codes
            Exact matches of product codes
        :param list or int product_code_start: list of product codes
            Partial matches of all products that start with this code
        :param list or str flow: list of flow, for example "import", "export"
        :param int period_start: filter on start period data
        :param int period_end: filter on end period data
        :return: A data frame of trade flows

        For example select monthly time series of Oak sawnwood trade

            >>> from biotrade.comtrade import comtrade
            >>> comtrade.db.select("monthly", product_code="440791")

        Select yearly time series of saw bean oil crude and other

            >>> comtrade.db.select("yearly", product_code=["150710", "150790"])

        """
        # Change character or integer arguments to lists suitable for a
        # column.in_() clause or for a list comprehension.
        if isinstance(reporter, str):
            reporter = [reporter]
        if isinstance(partner, str):
            partner = [partner]
        if isinstance(reporter_code, int):
            reporter_code = [reporter_code]
        if isinstance(partner_code, int):
            partner_code = [partner_code]
        if isinstance(product_code, str):
            product_code = [product_code]
        if isinstance(product_code_start, str):
            product_code_start = [product_code_start]
        if isinstance(flow, str):
            flow = [flow]
        # Select complementary tables
        reporter_table = self.tables["reporter"]
        partner_table = self.tables["partner"]
        quantity_table = self.tables["quantity_unit"]
        flow_table = self.tables["flow"]
        transport_table = self.tables["modality_of_transport"]
        custom_table = self.tables["custom"]
        if reporter is not None:
            # Obtain codes from reporter table
            stmt = reporter_table.select().where(
                reporter_table.c.reporter.in_(reporter)
            )
            # Define the new select table
            reporter_table = stmt.subquery()
        if partner is not None:
            # Obtain codes from partner table
            stmt = partner_table.select().where(partner_table.c.partner.in_(partner))
            # Define the new select table
            partner_table = stmt.subquery()
        if flow is not None:
            # Define a compiled regex
            regex_pat = re.compile(r"\W+")
            # Rename flow list to snake case using the compiled regex
            flow = [regex_pat.sub("_", item).lower() for item in flow]
            # Obtain codes from flow table
            stmt = flow_table.select().where(flow_table.c.flow.in_(flow))
            # Define the new select table
            flow_table = stmt.subquery()
        # Build the main select statement
        table = self.tables[table]
        stmt = table.select()
        if reporter_code is not None:
            stmt = stmt.where(table.c.reporter_code.in_(reporter_code))
        if partner_code is not None:
            stmt = stmt.where(table.c.partner_code.in_(partner_code))
        if product_code is not None:
            stmt = stmt.where(table.c.product_code.in_(product_code))
        if product_code_start is not None:
            stmt = stmt.where(
                or_(
                    table.c.product_code.ilike(f"{code}%")
                    for code in product_code_start
                )
            )
        if period_start is not None:
            stmt = stmt.where(table.c.period >= period_start)
        if period_end is not None:
            stmt = stmt.where(table.c.period <= period_end)
        # Subquery results
        table = stmt.subquery()
        # Join tables
        join_table = select(
            table,
            reporter_table.c.reporter,
            reporter_table.c.reporter_iso,
            partner_table.c.partner,
            partner_table.c.partner_iso,
            quantity_table.c.unit,
            flow_table.c.flow,
            transport_table.c.mode_of_transport,
            custom_table.c.custom_proc,
        ).select_from(
            table.join(
                reporter_table,
                table.c.reporter_code == reporter_table.c.reporter_code,
            )
            .join(
                partner_table,
                table.c.partner_code == partner_table.c.partner_code,
            )
            .join(
                quantity_table,
                table.c.unit_code == quantity_table.c.unit_code,
            )
            .join(
                flow_table,
                table.c.flow_code == flow_table.c.flow_code,
            )
            .join(
                transport_table,
                table.c.mode_of_transport_code
                == transport_table.c.mode_of_transport_code,
            )
            .join(
                custom_table,
                table.c.custom_proc_code == custom_table.c.custom_proc_code,
            )
        )
        # Query the database and return a data frame
        df = self.read_sql_query(join_table)
        return df

    def select_simple(
        self,
        table,
        reporter_code=None,
        partner_code=None,
        product_code=None,
        product_code_start=None,
        period_start=None,
        period_end=None,
    ):
        """
        Select comtrade trade data for the given arguments

        :param str table: name of the database table to select from
        :param list or int reporter_code: list of reporter codes
        :param list or int partner_code: list of partner codes
        :param list or int product_code: list of product codes
            Exact matches of product codes
        :param list or int product_code_start: list of product codes
            Partial matches of all products that start with this code
        :param int period_start: filter on start period data
        :param int period_end: filter on end period data
        :return: A data frame of trade flows

        For example select monthly time series of Oak sawnwood trade

            >>> from biotrade.comtrade import comtrade
            >>> comtrade.db.select_simple("monthly", product_code="440791")

        """
        # Change character or integer arguments to lists suitable for a
        # column.in_() clause or for a list comprehension.
        if isinstance(reporter_code, int):
            reporter_code = [reporter_code]
        if isinstance(partner_code, int):
            partner_code = [partner_code]
        if isinstance(product_code, str):
            product_code = [product_code]
        if isinstance(product_code_start, str):
            product_code_start = [product_code_start]
        # Build the main select statement
        table = self.tables[table]
        stmt = table.select()
        if reporter_code is not None:
            stmt = stmt.where(table.c.reporter_code.in_(reporter_code))
        if partner_code is not None:
            stmt = stmt.where(table.c.partner_code.in_(partner_code))
        if product_code is not None:
            stmt = stmt.where(table.c.product_code.in_(product_code))
        if product_code_start is not None:
            stmt = stmt.where(
                or_(
                    table.c.product_code.ilike(f"{code}%")
                    for code in product_code_start
                )
            )
        if period_start is not None:
            stmt = stmt.where(table.c.period >= period_start)
        if period_end is not None:
            stmt = stmt.where(table.c.period <= period_end)
        # Query the database and return a data frame
        df = self.read_sql_query(stmt)
        return df

    def select_long(self, *args, **kwargs):
        """Same as select but in long format with an element column and where
        the empty columns have been removed.

        Example:

            >>> from biotrade.comtrade import comtrade
            >>> swd = comtrade.db.select_long("monthly", product_code="440710")

        """
        df = self.select(*args, **kwargs)
        assert len(df) > 0
        # Remove empty columns
        not_empty = [col for col in df.columns if not all(df[col].isna())]
        df = df[not_empty]
        # This can be removed once this is dealt with in comtrade/pump.py
        # Rename column content to snake case using a compiled regex
        regex_pat = re.compile(r"\W+")
        if "flow" in df.columns:
            df["flow"] = df["flow"].str.replace(regex_pat, "_", regex=True).str.lower()
            # Remove the plural "s"
            df["flow"] = df["flow"].str.replace("s", "", regex=True)
        # TODO: Change period to a date time object
        # Reshape to long format
        value_cols = ["net_weight", "trade_value"]
        index = set(df.columns) - set(value_cols)
        df_long = df.melt(
            id_vars=list(index),
            value_vars=value_cols,
            var_name="element",
            value_name="value",
        )
        # Remove NA values for the weight
        df_long = df_long[~df_long.value.isna()]

        # Fix an issue with element appearing as a list in R
        # I don't understand why the element column is of sqlalchemy.sql.elements nature?
        # R> flatten(swd_con$element[1])
        # Error in `stop_bad_type()`:
        # ! Element 1 of `.x` must be a vector, not a `sqlalchemy.sql.elements.quoted_name
        # /sqlalchemy.util.langhelpers.MemoizedSlots/python.builtin.str/python.builtin.obj
        # ect` object
        df_long = df_long.reset_index(drop=True)
        df_long["element"] = df_long["element"].str.strip()
        return df_long


class DatabaseComtradePostgresql(DatabaseComtrade):
    """Database using the PostgreSQL engine use the same database for all data sources
    a separate schema for each data source

    The connection URL is defined in an environment variable, for example by
    adding the following to  your .bash_aliases or .bash_rc:

        export BIOTRADE_DATABASE_URL="postgresql://rdb@localhost/biotrade"
    """

    database_url = database_url
    schema = "raw_comtrade"


class DatabaseComtradeSqlite(DatabaseComtrade):
    """Database using the SQLite engine, use a separate database for each data source"""

    database_url = f"sqlite:///{data_dir}/comtrade/comtrade.db"
    schema = "main"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Paul Rougieux and Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

JRC biomass Project.
Unit D1 Bioeconomy.

Read forestry production and trade data from the database

    >>> from biotrade.faostat import faostat

Read forestry production and trade data from the bulk FAOSTAT files directly
(loading large files takes a few seconds)

    >>> fp = faostat.pump.read_df("forestry_production")
    >>> ft = faostat.pump.read_df("forestry_trade")

Download or update zipped CSV files from FAOSTAT.

    >>> from biotrade.faostat import faostat
    >>> faostat.pump.download_zip_csv("Forestry_E_All_Data_(Normalized).zip")

Display information on column names used for renaming
and dropping less important columns:

    >>> faostat.column_names

Display the list of EU countries names in the FAOSTAT data

    >>> faostat.country_groups.eu_country_names

"""
# Third party modules
import pandas

# Internal modules
from biotrade import module_dir, data_dir, database_url
from biotrade.faostat.coefficients import Coefficients
from biotrade.faostat.country_groups import CountryGroups
from biotrade.faostat.database import DatabaseFaostatPostgresql
from biotrade.faostat.database import DatabaseFaostatSqlite
from biotrade.faostat.products import Products
from biotrade.faostat.pump import Pump

# Define a logging mechanism to keep track of errors and debug messages
from biotrade.common.logger import create_logger

create_logger()


class Faostat:
    """
    Parent to various objects dealing with FAOSTAT data
    """

    # Location of the data
    data_dir = data_dir / "faostat"

    # Location of module configuration data
    config_data_dir = module_dir / "config_data"

    # Load a mapping table used to rename columns
    df = pandas.read_csv(config_data_dir / "column_names.csv")
    df = df.filter(regex="biotrade|faostat")
    non_na_values = (~df.filter(like="faostat").isna()).sum(axis=1)
    column_names = df[non_na_values > 0]

    def __init__(self):
        # Create a faostat subfolder in the data directory if it doesn't exist
        (data_dir / "faostat").mkdir(exist_ok=True)
        if not self.data_dir.exists():
            self.data_dir.mkdir()

    @property
    def country_groups(self):
        """Identify reporter and partner countries and regions"""
        return CountryGroups(self)

    @property
    def products(self):
        """Identify Comtrade products (commodities) and metadata"""
        return Products(self)

    @property
    def pump(self):
        """Load data from FAOSTAT and read it into data frames"""
        return Pump(self)

    @property
    def coefficients(self):
        """Coefficients specific to FAOSTAT"""
        return Coefficients(self)

    @property
    def db_sqlite(self):
        """Database using the SQLite engine"""
        return DatabaseFaostatSqlite(self)

    @property
    def db_postgresql(self):
        """Database using the PostGreSQL engine"""
        return DatabaseFaostatPostgresql(self)

    @property
    def db(self):
        """The generic database can be either a PostGreSQL or a SQLite database
        Depending of the value of the environmental variable
        BIOTRADE_DATABASE_URL. If it's not defined use an SQLite database, otherwise
        use a PostGreSQL database as defined in that URL. Note that the environment
        variables are read at the root of this module's directory. In
        particular BIOTRADE_DATABASE_URL is stored into the DATABASE_URL
        variable.
        """
        if database_url is None:
            return DatabaseFaostatSqlite(self)
        return DatabaseFaostatPostgresql(self)


# Make a singleton #
faostat = Faostat()



"""
Written by Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

Script which contains functions used to produce data for the Web Platform

"""

import os
import warnings
import pandas as pd
import numpy as np
from pathlib import Path
from biotrade.faostat import faostat
from biotrade import data_dir
from biotrade.common.compare import merge_faostat_comtrade, ELEMENT_DICT
from biotrade.common.time_series import (
    relative_absolute_change,
    segmented_regression,
    merge_analysis,
)

# Define column name suffixes for average, total and percentage calculations
COLUMN_AVG_SUFFIX = "_avg_value"
COLUMN_PERC_SUFFIX = "_percentage"
COLUMN_TOT_SUFFIX = "_tot_value"


def filter_trade_data(df):
    """
    Filter trade data and change units
    """
    # Remove trade products where unit is different from kg and usd
    df = df[df["element_code"].isin(ELEMENT_DICT["element_code"])].reset_index(
        drop=True
    )
    element_df = pd.DataFrame.from_dict(ELEMENT_DICT)
    # Convert trade values from USD to Million USD
    codes = element_df[element_df["element"].str.endswith("_value")][
        "element_code"
    ].values.tolist()
    selector = df["element_code"].isin(codes)
    df.loc[selector, "value"] = df.loc[selector, "value"] * 1e-6
    df.loc[selector, "unit"] = "Mdollar"
    # Convert kg to tonnes
    codes = element_df[element_df["element"].str.endswith("_quantity")][
        "element_code"
    ].values.tolist()
    selector = df["element_code"].isin(codes)
    df.loc[selector, "value"] = df.loc[selector, "value"] * 1e-3
    df.loc[selector, "unit"] = "ton"
    return df


def filter_production_data(df):
    """
    Filter crop data and change units
    """
    # All units in lower case
    df["unit"] = df["unit"].str.lower()
    units_allowed = ["100 g/an", "100 g/ha", "an", "ha", "m3", "t"]
    # Put control on unit names
    if sorted(df["unit"].unique()) != units_allowed:
        warnings.warn(
            f"Units of production dataset have been changed from\n{units_allowed}\nto\n{sorted(df['unit'].unique())}"
        )
    # Remove yields for animals
    df = df[df["unit"] != "100 g/an"]
    # Use ton as unit name for tonnes
    selector = df["unit"] == "t"
    df.loc[selector, "unit"] = "ton"
    # Use head as unit name for animal
    selector = df["unit"] == "an"
    df.loc[selector, "unit"] = "head"
    # Yield data transformed from hg/ha into ton/ha
    selector = df["element"] == "yield"
    df.loc[selector, "value"] = df.loc[selector, "value"] / 10**4
    df.loc[selector, "unit"] = "ton/ha"
    # Select only at country level
    df = df[df["reporter_code"] < 1000].reset_index(drop=True)
    return df


def replace_zero_with_nan_values(df, column_list):
    """
    Replace zero value columns with nan, to let the values be dropped afterwards

    :param df (DataFrame), output to be saved
    :param column_list (list), name of columns to replace 0 with nan values
    :return df (DataFrame), with nan replacements

    """
    df[column_list] = df[column_list].replace(0, np.nan)
    return df


def save_file(df, file_name):
    """
    Function which save output scripts

    :param df (DataFrame), output to be saved
    :param file_name (str), name of the output file

    """
    # Save csv files to env variable path or into biotrade data folder alternatively
    if os.environ.get("FRONT_END_DATA"):
        path = Path(os.environ["FRONT_END_DATA"])
    else:
        path = data_dir / "front_end"
    path.mkdir(exist_ok=True)
    df.to_csv(path / file_name, index=False, na_rep="null")
    print(f"Saved file {file_name} to {path}")


def comtrade_products():
    """
    Return the regulation product codes and names together with the associated 6 digit codes of Comtrade products and Faostat contained into the file biotade/config_data/regulation_products.csv

    :return df (Dataframe), containing Comtrade and Faostat product codes and names
    """
    # Name of product file to retrieve
    main_product_file = faostat.config_data_dir / "regulation_products.csv"
    # Retrieve dataset
    df = pd.read_csv(
        main_product_file,
        dtype={"regulation_code": str, "hs_4d_code": str, "hs_6d_code": str},
    )
    # Retrieve name and codes and return the dataframe
    columns = [
        "regulation_code",
        "regulation_short_name",
        "commodity_short_name",
        "hs_6d_code",
        "fao_code",
        "match",
    ]
    df = df[columns]
    df.rename(
        columns={
            "regulation_code": "product_code",
            "regulation_short_name": "product_name",
            "commodity_short_name": "commodity_name",
            "hs_6d_code": "comtrade_code",
        },
        inplace=True,
    )
    return df


def main_product_list(table_list):
    """
    Return the main list of Faostat products (without duplicates) contained into the file biotade/config_data/faostat_products_name_code_shortname.csv
    depending on the purpose: production or trade

    :parameter table_list (list), list of the tables to retrieve product codes
    :return product_list (list), list of the main Faostat product codes

    """
    # Name of product file to retrieve
    main_product_file = (
        faostat.config_data_dir / "faostat_products_name_code_shortname.csv"
    )
    # Retrieve dataset of regulation products
    main_products = pd.read_csv(main_product_file)
    # Select only product codes with an associated commodity from the commodity datataset
    main_products = (
        main_products[~main_products.commodity_name.isnull()]
        .code.drop_duplicates()
        .to_list()
    )
    # Define db and pre allocate dataframe
    db = faostat.db
    df = pd.DataFrame(columns=["product_code"])
    # Define which products are inside the production/trade list
    for table in table_list:
        table = db.tables[table]
        df_table = pd.read_sql_query(
            table.select()
            .distinct(table.c.product_code)
            .with_only_columns([table.c.product_code]),
            db.engine,
        )
        df = pd.concat([df, df_table], ignore_index=True)
    # Drop db product duplicates
    product_list = df.product_code.drop_duplicates().to_list()
    # Obtain the intersection
    product_list = list(set(main_products).intersection(product_list))
    return product_list


def reporter_iso_codes(df):
    """
    Script which transforms faostat reporter and partner codes into iso3 codes

    :param df (DataFrame), which contains reporter_code and (if applicable) partner_code columns
    :return df (DataFrame), with the substitution into iso3 codes

    """
    # Reporter codes
    reporter_file = faostat.config_data_dir / "faostat_country_groups.csv"
    reporter = pd.read_csv(reporter_file)
    # Obtain iso3 codes for reporters and partners
    df = df.merge(
        reporter[["faost_code", "iso3_code"]],
        how="left",
        left_on="reporter_code",
        right_on="faost_code",
    )
    df["reporter_code"] = df["iso3_code"]
    if "partner_code" in df.columns:
        df.drop(columns=["faost_code", "iso3_code"], inplace=True)
        df = df.merge(
            reporter[["faost_code", "iso3_code"]],
            how="left",
            left_on=["partner_code"],
            right_on=["faost_code"],
        )
        df["partner_code"] = df["iso3_code"]
    df.drop(columns=["faost_code", "iso3_code"], inplace=True)
    # Consider only data of official country codes by GISCO
    country_codes = (
        pd.read_csv(
            data_dir / "GISCO_CNTR_LIST.txt",
            sep=";",
        )
        .ISO3_CODE.drop_duplicates()
        .to_list()
    )
    df = df[df.reporter_code.isin(country_codes)].reset_index(drop=True)
    if "partner_code" in df.columns:
        df = df[df.partner_code.isin(country_codes)].reset_index(drop=True)
        # Remove free zones internal trade data
        df = df[df["reporter_code"] != df["partner_code"]].reset_index(drop=True)
    return df


def merge_faostat_comtrade_data(
    faostat_code=None,
    comtrade_regulation=None,
    aggregate=True,
    nr_products_chunk=10,
):
    """
    Script which merges faostat yearly trade data with comtrade yearly and monthly trade data.
    The last 12 monthly comtrade data are aggregated to reconstruct the most recent year.
    Considered only import and export data. Re-import and re-export excluded.

    :param faostat_code (list), list of product codes to be retrieved
    :param dataframe comtrade_regulation: comtrade regulation codes to be loaded and aggregated, default is None and in this case they are mapped into faostat codes
    :param boolean aggregate: data are aggregated or not by product code, default is True
    :param int nr_products_chunk: split regulation codes into chucks to avoid run out of memory
    :return df_merge (DataFrame), dataframe with merged data

    """
    # Pre allocate dataframe
    df_merge = pd.DataFrame()
    # Select quantities from Faostat db and Comtrade for trade data for all countries (code < 1000)
    if comtrade_regulation is not None:
        regulation_products = comtrade_regulation.product_code.unique().tolist()
    else:
        regulation_products = [None]
    index_list = [
        "source",
        "reporter_code",
        "reporter",
        "partner_code",
        "partner",
        "product_code",
        "product",
        "element_code",
        "element",
        "year",
        "unit",
    ]
    # Split product list in chunks to avoid run out of memory
    for i in range(0, len(regulation_products), nr_products_chunk):
        regulation_code = regulation_products[i : i + nr_products_chunk]
        if comtrade_regulation is not None:
            comtrade_code = comtrade_regulation[
                comtrade_regulation.product_code.isin(regulation_code)
            ].comtrade_code.to_list()
        else:
            comtrade_code = None
        df = merge_faostat_comtrade(
            "crop_trade",
            "yearly",
            faostat_code,
            comtrade_code,
            aggregate,
        )
        # Select only import and export (exclude re-import and re-export for now) and remove nan reporters/partners (=-1)
        df = df[
            (
                df.element.isin(
                    [
                        "import_value",
                        "import_quantity",
                        "export_value",
                        "export_quantity",
                    ]
                )
            )
            & (~(df[["reporter_code", "partner_code"]] == -1).any(axis=1))
        ].reset_index(drop=True)
        if comtrade_regulation is not None:
            # Merge to obtain regulation product codes
            df = df.merge(
                comtrade_regulation,
                left_on="product_code",
                right_on="comtrade_code",
                how="left",
                suffixes=("", "_regulation"),
            )
            # Replace comtrade products with product code regulations to aggregate
            selector = df.source == "comtrade"
            df.loc[selector, "product"] = df.loc[selector, "product_name"]
            df.loc[selector, "product_code"] = df.loc[
                selector, "product_code_regulation"
            ]
            df = df.groupby(index_list)["value"].agg("sum").reset_index()
        df_merge = pd.concat([df_merge, df], ignore_index=True)
        # Avoid to retrieve for all the cycle the same faostat data
        if i == 0:
            faostat_code = None
    # Use period instead of year column
    df_merge["period"] = df_merge["year"]
    return df_merge


def average_results(df, threshold, dict_list, interval_array=np.array([])):
    """
    Script which produce the average and percentage results for the tree maps of the web platform

    :param df (DataFrame), database to perform calculations
    :param threshold (int), percentage above which classifying as "Others" the percentages in the tree maps
    :param dict_list (list), list of dictionaries containing the column names for the aggregation and the percentages, as well the code for the "Others" category and possible columns to be added to the group by aggregation
    :param interval_array (array), array with incremental weights (between 0 and 1) to assign data intervals. If empty (default), intervals are not calculated
    :return df_final (DataFrame), where calculations are performed

    """
    # Make a copy of df to avoid overrides
    df = df.copy()
    # Query df to obtain most recent year of data
    most_recent_year = sorted(df.year.unique(), reverse=True)[0]
    # Consider all the years of the table except the most recent (no aggregation for it)
    years = sorted(df.year.unique(), reverse=True)[1:]
    # Define aggregation of 5 years at a time, starting from the most recent year - 1
    periods = np.array(range(0, len(years))) // 5 + 1
    # Construct the df_periods containing the periods yyyy-yyyy and merge it with df
    df_periods = pd.DataFrame(
        {
            "year": [most_recent_year, *years],
            "period_aggregation": [0, *periods],
        }
    )
    # Define the min year inside each period
    df_periods_min = (
        df_periods.groupby(["period_aggregation"]).agg({"year": "min"}).reset_index()
    )
    # Define the max year inside each period
    df_periods_max = (
        df_periods.groupby(["period_aggregation"]).agg({"year": "max"}).reset_index()
    )
    # Merge on the periods
    df_periods = df_periods.merge(
        df_periods_min,
        how="left",
        on="period_aggregation",
        suffixes=("", "_min"),
    )
    df_periods = df_periods.merge(
        df_periods_max,
        how="left",
        on="period_aggregation",
        suffixes=("", "_max"),
    )
    # Define the structure yyyy-yyyy for the period aggregation column
    df_periods["period_aggregation"] = (
        df_periods["year_min"].astype(str) + "-" + df_periods["year_max"].astype(str)
    )
    # Assign the associated period to each data
    df = df.merge(df_periods[["year", "period_aggregation"]], on="year", how="left")
    df["period"] = df["period_aggregation"]
    # Default index list for aggregations + the adds from the arguments
    index_list = ["element", "period", "unit"]
    df_final = pd.DataFrame()
    column_drop = []
    # Load for each dict the aggregation column, the percentage column and the code for threshold values in order to compute calculations
    for dict in dict_list:
        index_list_upd = [
            dict["average_col"],
            *index_list,
            *dict["index_list_add"],
        ]
        value_col_name = "value"
        average_col_name = dict["average_col"] + COLUMN_AVG_SUFFIX
        total_col_name = dict["average_col"] + COLUMN_TOT_SUFFIX
        percentage_col_name = dict["percentage_col"] + COLUMN_PERC_SUFFIX
        other_code = dict["threshold_code"]
        # For the aggregation column calculate mean and sum of the values in the given period aggregation, related to a specific unit and element
        df_mean = (
            df.groupby([*index_list_upd, "year"])
            .agg({"value": "sum"})
            .reset_index()
            .groupby(index_list_upd)
            .agg({"value": "mean"})
            .reset_index()
            .rename(columns={"value": average_col_name})
        )
        df_total = (
            df.groupby(index_list_upd)
            .agg({"value": "sum"})
            .reset_index()
            .rename(columns={"value": total_col_name})
        )
        if total_col_name not in column_drop:
            column_drop.append(total_col_name)
        # Percentage column aggregated with the column list
        df_new = (
            df.groupby([*index_list_upd, dict["percentage_col"]])
            .agg({"value": "sum"})
            .reset_index()
        )
        if value_col_name not in column_drop:
            column_drop.append(value_col_name)
        # Merge with mean and total values
        df_new = df_new.merge(df_mean, how="left", on=index_list_upd)
        df_new = df_new.merge(df_total, how="left", on=index_list_upd)
        # Percentage associated to the percentage column on a given aggregated period
        df_new[percentage_col_name] = df_new["value"] / df_new[total_col_name] * 100
        # Sort by percentage, compute the cumulative sum and shift it by one
        df_new.sort_values(
            by=[*index_list_upd, percentage_col_name],
            ascending=False,
            inplace=True,
            ignore_index=True,
        )
        # Skip nan values is True for cumsum by default
        df_new["cumsum"] = df_new.groupby(index_list_upd)[percentage_col_name].cumsum()
        df_new["cumsum_lag"] = df_new.groupby(index_list_upd)["cumsum"].transform(
            "shift", fill_value=0
        )
        # Create a grouping variable instead of the percentage column, which will be 'Others' for
        # values above the threshold
        df_new[dict["percentage_col"]] = df_new[dict["percentage_col"]].where(
            df_new["cumsum_lag"] < threshold, other_code
        )
        # Group the percentage column values which are in the 'Others' category and calculate their percentage
        df_new = (
            df_new.groupby([*index_list_upd, dict["percentage_col"]])
            .agg(
                {
                    "value": "sum",
                    average_col_name: "first",
                    total_col_name: "first",
                }
            )
            .reset_index()
        )
        df_new[percentage_col_name] = df_new["value"] / df_new[total_col_name] * 100
        if df_final.empty:
            df_final = df_new
        else:
            # Merge or concatenate depending on the common columns in merge_col_list
            merge_col_list = [*index_list_upd, dict["percentage_col"]]
            if average_col_name in df_final.columns:
                merge_col_list.append(average_col_name)
            # Meaning that dataframes can be merged
            if set(merge_col_list).issubset(df_final.columns):
                merge_suffix = "_merge"
                df_final = df_final.merge(
                    df_new,
                    how="outer",
                    on=merge_col_list,
                    suffixes=("", merge_suffix),
                )
                # Add merge columns to column drop list
                for col in df_final.columns:
                    if col.endswith(merge_suffix):
                        if col not in column_drop:
                            column_drop.append(col)
            else:
                df_final = pd.concat([df_final, df_new], ignore_index=True)
    # Remove columns from df_final
    df_final.drop(columns=column_drop, inplace=True)
    # Compute avg production for a certain commodity, reporter, (partner) and period
    groupby_avg_cols = [
        "product_code",
        "element",
        "unit",
        "reporter_code",
        "period",
    ]
    if "partner_code" in df.columns:
        groupby_avg_cols.append("partner_code")
    # Calculate the average over time
    df_avg = (
        df.groupby(groupby_avg_cols)
        .agg({"value": "mean"})
        .reset_index()
        .rename(columns={"value": "avg_value"})
    )
    if len(interval_array):
        # Extract max value avg production for a commodity across periods and countries
        groupby_max_cols = ["product_code", "element", "unit"]
        df_max = (
            df_avg.groupby(groupby_max_cols)
            .agg({"avg_value": "max"})
            .reset_index()
            .rename(columns={"avg_value": "max_avg_value"})
        )
        df_avg = df_avg.merge(df_max, how="left", on=groupby_max_cols)
        # Compute thresholds: [interval_array] * df["max_avg_value"]
        df_avg["bin"] = (
            df_avg["max_avg_value"].values.reshape(len(df_avg), 1)
            * np.array([interval_array])
        ).tolist()
        # For each group (product, element, unit) define to which interval the average production of the specific country and period belongs
        df_groups = df_avg.groupby(groupby_max_cols)
        df_avg = pd.DataFrame()
        df_legend = pd.DataFrame()
        for key in df_groups.groups.keys():
            df_key = df_groups.get_group(key).reset_index(drop=True)
            key_legend = pd.DataFrame()
            bins = df_key.bin[0]
            # If no max_avg_production defined, then put nan
            if df_key.max_avg_value.isnull().all():
                df_key["interval"] = pd.cut(
                    df_key["avg_value"],
                    bins=bins,
                    duplicates="drop",
                )
            # Assign the intervals and ranges
            elif len(df_key.max_avg_value.unique()) == 1:
                df_key["interval"] = pd.cut(
                    df_key["avg_value"],
                    bins=bins,
                    labels=list(range(0, len(interval_array) - 1)),
                )
                df_key["interval_range"] = pd.cut(
                    df_key["avg_value"],
                    bins=bins,
                )
                key_legend["interval"] = df_key["interval"].cat.categories.values
                key_legend["min_value"] = df_key["interval_range"].cat.categories.left
                key_legend["max_value"] = df_key["interval_range"].cat.categories.right
                key_legend["product_code"] = key[0]
                key_legend["element"] = key[1]
                key_legend["unit"] = key[2]
            # Inconsistencies could be detected with the warning
            else:
                faostat.db.logger.warning(
                    f"The dataset represented by {groupby_max_cols} = {key} tuple has not been included into the final csv file due to inconsistencies, please check the data."
                )
                continue
            # Concat to obtain updated dataframes
            df_legend = pd.concat(
                [df_legend, key_legend],
                ignore_index=True,
            )
            df_avg = pd.concat(
                [df_avg, df_key[[*groupby_avg_cols, "avg_value", "interval"]]],
                ignore_index=True,
            )
        # Columns to keep in the legend dataframe
        drop_column = "element"
        column_list = df_legend.columns.tolist()
        column_list.remove(drop_column)
        # Save interval legends
        harvested_area_legend = df_legend[df_legend["element"] == "area_harvested"][
            column_list
        ]
        # Put legend in million hectares
        selector = harvested_area_legend.interval == 0
        harvested_area_legend.loc[selector, "description"] = "up to " + (
            harvested_area_legend.loc[selector, "max_value"] / 10**6
        ).round(2).astype(str)
        harvested_area_legend.loc[~selector, "description"] = (
            "from "
            + (harvested_area_legend.loc[~selector, "min_value"] / 10**6)
            .round(2)
            .astype(str)
            + " to "
            + (harvested_area_legend.loc[~selector, "max_value"] / 10**6)
            .round(2)
            .astype(str)
        )
        harvested_area_legend["description"] = (
            harvested_area_legend["description"] + " M" + harvested_area_legend.unit
        )
        save_file(harvested_area_legend, "harvested_area_average_legend.csv")
        production_legend = df_legend[
            df_legend["element"].isin(["production", "stocks"])
        ][column_list]
        # Put legend in Million or kilo (for products 839 and 869) tonnes, m3 and heads
        selector = (production_legend.interval == 0) & (
            production_legend.product_code.isin([839, 869])
        )
        production_legend.loc[selector, "description"] = (
            "up to "
            + (production_legend.loc[selector, "max_value"] / 10**3)
            .round(2)
            .astype(str)
            + " k"
        )
        selector = (production_legend.interval == 0) & ~(
            production_legend.product_code.isin([839, 869])
        )
        production_legend.loc[selector, "description"] = (
            "up to "
            + (production_legend.loc[selector, "max_value"] / 10**6)
            .round(2)
            .astype(str)
            + " M"
        )
        selector = (production_legend.interval != 0) & (
            production_legend.product_code.isin([839, 869])
        )
        production_legend.loc[selector, "description"] = (
            "from "
            + (production_legend.loc[selector, "min_value"] / 10**3)
            .round(2)
            .astype(str)
            + " to "
            + (production_legend.loc[selector, "max_value"] / 10**3)
            .round(2)
            .astype(str)
            + " k"
        )
        selector = (production_legend.interval != 0) & ~(
            production_legend.product_code.isin([839, 869])
        )
        production_legend.loc[selector, "description"] = (
            "from "
            + (production_legend.loc[selector, "min_value"] / 10**6)
            .round(2)
            .astype(str)
            + " to "
            + (production_legend.loc[selector, "max_value"] / 10**6)
            .round(2)
            .astype(str)
            + " M"
        )
        production_legend["description"] = (
            production_legend["description"] + production_legend["unit"]
        )
        save_file(production_legend, "production_average_legend.csv")
        # Save yield interval legends
        yield_legend = df_legend[df_legend["element"] == "yield"][column_list]
        # Define two statements
        selector = yield_legend.interval == 0
        yield_legend.loc[selector, "description"] = "up to " + (
            yield_legend.loc[selector, "max_value"]
        ).round(2).astype(str)
        yield_legend.loc[~selector, "description"] = (
            "from "
            + (yield_legend.loc[~selector, "min_value"]).round(2).astype(str)
            + " to "
            + (yield_legend.loc[~selector, "max_value"]).round(2).astype(str)
        )
        yield_legend["description"] = (
            yield_legend["description"] + " " + yield_legend.unit
        )
        save_file(yield_legend, "yield_average_legend.csv")
    # Associate the avg productions (eventually with intervals) to the final dataframe
    df_final = df_final.merge(df_avg, on=groupby_avg_cols, how="left")
    return df_final


def aggregated_data(
    df,
    code_list,
    agg_country_code,
    agg_country_name,
    groupby_cols=[
        "product_code",
        "product",
        "element_code",
        "element",
        "year",
        "unit",
    ],
    value_cols=["value"],
):
    """
    Script that aggregates country data in order to overcome inconsistencies

    :param df (DataFrame), which contains production or trade data
    :param code_list (list), country codes to be aggregated
    :param agg_country_code (int), country code to be assigned for the aggregation
    :param agg_country_name (string), country name to be assigned for the aggregation
    :param groupby_cols (list), columns to consider for the groupby functions
    :param value_cols (list), columns on which to calculate the sum

    :return df (DataFrame), with aggregated data

    """
    # Trade data
    if "partner_code" in df.columns:
        # Define aggregated data both for reporter and partner
        df_agg = df[df[["reporter_code", "partner_code"]].isin(code_list).any(axis=1)]
        # Avoid counting internal trades
        mask = (df_agg.reporter_code.isin(code_list)) & (
            df_agg.partner_code.isin(code_list)
        )
        df_agg = df_agg[~mask]
        # Remove country code list data from df dataset
        df = df[~(df[["reporter_code", "partner_code"]].isin(code_list)).any(axis=1)]
        # Aggregation on the reporter side
        df_agg_1 = (
            df_agg[df_agg["reporter_code"].isin(code_list)]
            .groupby(
                [
                    *groupby_cols,
                    "source",
                    "partner_code",
                    "partner",
                ]
            )
            # If all null values, do not return 0 but Nan
            .agg({col: lambda x: x.sum(min_count=1) for col in value_cols})
            .reset_index()
        )
        df_agg_1["reporter_code"] = agg_country_code
        df_agg_1["reporter"] = agg_country_name
        # Concat with reporter codes not in country code list
        df_agg_1 = pd.concat(
            [df_agg[~df_agg["reporter_code"].isin(code_list)], df_agg_1],
            ignore_index=True,
        )
        # Aggregation also on the partner side
        df_agg_2 = (
            df_agg_1[df_agg_1["partner_code"].isin(code_list)]
            .groupby(
                [
                    *groupby_cols,
                    "source",
                    "reporter_code",
                    "reporter",
                ]
            )
            # If all null values, do not return 0 but Nan
            .agg({col: lambda x: x.sum(min_count=1) for col in value_cols})
            .reset_index()
        )
        df_agg_2["partner_code"] = agg_country_code
        df_agg_2["partner"] = agg_country_name
        # Concat with partner codes not in country code list
        df_agg_2 = pd.concat(
            [
                df_agg_1[~df_agg_1["partner_code"].isin(code_list)],
                df_agg_2,
            ],
            ignore_index=True,
        )
        df_agg = df_agg_2
    # Production data
    else:
        # Produce reporter aggregated data
        df_agg = df[df["reporter_code"].isin(code_list)]
        # Remove country code list data from df dataset
        df = df[~(df["reporter_code"].isin(code_list))]
        df_agg = (
            df_agg.groupby([*groupby_cols])
            # If all null values, do not return 0 but Nan
            .agg(
                {col: lambda x: x.sum(min_count=1) for col in value_cols}
            ).reset_index()
        )
        df_agg["reporter_code"] = agg_country_code
        df_agg["reporter"] = agg_country_name
    # Fill period column
    df_agg["period"] = df_agg["year"]
    # Build the final dataset to return
    df = pd.concat(
        [df, df_agg],
        ignore_index=True,
    )
    return df


def trend_analysis(
    df_data, multi_process=False, groupby_column_list=[], value_column=None
):
    """
    Script which performs the trend analysis

    :param df_data (DataFrame), data on which perform the analysis
    :param multi_process (Boolean), if True segmented regression is performed through multiple cores. Default is False
    :param groupby_column_list (List), columns to be grouped. Default is empty list
    :param value_column (string), column which refers to data. Default is None
    :return df (DataFrame), dataframe containing the relative, absolute and segmented regresssion indicators

    """
    # Calculate the absolute and relative change
    df_data_change = relative_absolute_change(
        df_data,
        last_value=True,
        groupby_column_list=groupby_column_list,
        value_column=value_column,
        multi_process=multi_process,
    )
    # Use as objective function the coefficient of determination (R2), significance level of 0.05 and at least 7 points for the linear regression
    df_data_regression = segmented_regression(
        df_data,
        last_value=True,
        function="R2",
        alpha=0.05,
        min_data_points=7,
        multi_process=multi_process,
        groupby_column_list=groupby_column_list,
        value_column=value_column,
    )
    # Merge dataframes to compare results
    df = merge_analysis(
        df_change=df_data_change, df_segmented_regression=df_data_regression
    )
    # Define the structure yyyy-yyyy for the period change and segmented regression columns
    df["period_change"] = (
        df["year_range_lower_change"].astype("Int64").astype(str)
        + "-"
        + df["year_range_upper_change"].astype("Int64").astype(str)
    ).replace("<NA>-<NA>", np.nan)
    df["period_regression"] = (
        df["year_range_lower_regression"].astype("Int64").astype(str)
        + "-"
        + df["year_range_upper_regression"].astype("Int64").astype(str)
    ).replace("<NA>-<NA>", np.nan)
    # Transform from boolean to integers 0 or 1
    df["mk_significance_flag"] = (
        df["mk_ha_test"].astype("boolean").astype("Int64")
    ).replace(pd.NA, np.nan)
    # Transform unit in unit / year
    df["unit"] = df["unit"] + "/year"
    return df



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Paul Rougieux and Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

JRC biomass Project.
Unit D1 Bioeconomy.

Get the list of products as stored in the database

    >>> from biotrade.comtrade import comtrade
    >>> hs = comtrade.products.hs
    >>> # Refresh the HS product list by downloading it again from the Comtrade API
    >>> comtrade.pump.update_db_parameter()

Search by product names

    >>> # Paper related products
    >>> print(hs[hs.product_description.str.contains("paper", case=False)])
    >>> # The search pattern is a regex by default, so you can search
    >>> # for cocoa or chocolate related products with
    >>> print(hs[hs.product_description.str.contains("cocoa|choco", case=False)])

Search for products starting with a specific code

    >>> print(hs[hs.product_code.str.startswith("4403")])
    >>> # Write to a csv file
    >>> hs44 = hs[hs.product_code.str.startswith("44")]
    >>> hs44.to_csv("/tmp/hs_product_codes.csv", index=False)

Get the module internal list of products at the HS 2 digit level, with an
indicator of what products are saved when updating data. Where the column
bioeconomy == 1 the product is saved, bioeconomy == 0 the product is discarded.

    >>> products2d = comtrade.products.hs2d
    >>> # Paper related products
    >>> print(products2d[products2d.product_description.str.contains("paper")])

"""
# Third party modules
import pandas


class Products(object):
    """
    Comtrade product list, with additional information.

    Update the list of product codes from Comtrade

        >>> from biotrade.comtrade import comtrade
        >>> comtrade.pump.update_db_parameter()

    """

    def __init__(self, parent):
        self.parent = parent
        self.db = self.parent.db
        self.config_data_dir = self.parent.config_data_dir
        self.hs_csv_file = self.parent.data_dir / "classificationHS.csv"

    @property
    def hs(self):
        """List of products as stored in the database

        Usage:

            >>> from biotrade.comtrade import comtrade
            >>> comtrade.products.hs

        Refresh the HS product list by downloading it again from the Comtrade API

            >>> comtrade.pump.update_db_parameter()

        The list is downloaded from the Comtrade API and stored into the database.
        """
        df = self.db.select("product")
        return df

    @property
    def short_names(self):
        """List of short product names as stored in the package config_data folder

        Usage:

            >>> from biotrade.comtrade import comtrade
            >>> comtrade.products.short_names

        """
        file_name = self.config_data_dir / "comtrade_hs_product_short_names.csv"
        df = pandas.read_csv(
            file_name,
            dtype={"product_code": "str"},
        )
        dup = df.product_code.duplicated(keep=False)
        if any(dup):
            raise ValueError(
                f"Error in \n{file_name}\n"
                + f"The following codes are duplicated:\n{df[dup]}"
            )
        return df

    @property
    def hs2d(self):
        """The module internal list of products at the 2 digits level.

        Usage:

            >>> from biotrade.comtrade import comtrade
            >>> comtrade.products.hs2d

        For information the internal list is a filtered version of
        the list of Harmonized System (HS) commodities
        originally downloaded from Comtrade with the method:

            >>> comtrade.pump.get_parameter_list("classificationHS.json")

        """
        df = pandas.read_csv(
            self.config_data_dir / "comtrade_hs_2d.csv",
            # Force the product_code column to remain a character column,
            # otherwise str "01" becomes int 1.
            dtype={"product_code": str},
        )
        return df

    def under(self, product_code):
        """Give all products under the given product code"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Paul Rougieux and Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

JRC biomass Project.
Unit D1 Bioeconomy.

The `biotrade` package analyses international trade of bio-based products. It focuses on
the agriculture and forestry sectors, from primary production to secondary products
transformation. It loads bilateral trade data from UN Comtrade, production and trade
data from FAOSTAT and socio-economic indicators from the World Bank.

- For installation instructions see the
[README](https://pypi.org/project/biotrade/) on pypi.org or on the project git
repository.

The `biotrade` package is structured into sub-packages, one for each data
source: `biotrade.faostat`, `biotrade.comtrade` and `biotrade.world_bank`. In
addition, there is a `biotrade.common` sub-package.


# Update data from sources

Each data source has a "pump" and a "database" object. The "pump" loads data
from the source and the database stores data locally. The documentation of the
pump and database of each source provide information on how to update and
select data. Start with the FAOSTAT database which is the most light weight and
which will work correctly also without an API access token.

- `biotrade.faostat.pump` and `biotrade.faostat.database`

- `biotrade.comtrade.pump` and `biotrade.comtrade.database`

- `biotrade.world_bank.pump` and `biotrade.world_bank.database`


# Transform data

- Common data transformation functions are located in `biotrade.common`.

- Some transformation functions are specific to a data sources, you will find
the under the sub-module for that data source such as for example:
`biotrade.faostat.aggregate` or `biotrade.faostat.country_groups`


# Paths defined for biotrade data

Some paths are defined at the top level:

 - `module dir` makes it possible to access configuration data such as product
 codes and country codes mapping tables in `biotrade/config data`.

 - `data_dir` contains the location of the directory where data is stored (if
   SQLite databases are used, they will be stored there as well). By default it
   is located in the user's home directory, in a sub-directory called
   "$HOME/repos/forobs/biotrade_data/" but this can be changed with the
   environment variable BIOTRADE_DATA. To check that you have entered the
   environment variable correctly, you can call the following from python:

        >>> import os
        >>> os.environ["BIOTRADE_DATA"]

 - For more information on how to set up environment variables in Windows, Mac
   or Unix see
   https://superuser.com/questions/284342/what-are-path-and-other-environment-variables-and-how-can-i-set-or-use-them

 - `database_url` is equal to `None` by default and then a local on-disk SQLite
 database is used. It can be changed to a PostGreSQL database through the
 environment variable BIOTRADE_DATABASE_URL that specifies the location of the
 database connection URL.

 You can display these values with :

    >>> import biotrade
    >>> print("module_dir:", biotrade.module_dir)
    >>> print("data_dir:", biotrade.data_dir)
    >>> print("database_url:", biotrade.database_url)

"""

from pathlib import Path
import os

__version__ = "0.3.16"

module_dir = Path(__file__).resolve().parent

# Where is the data, default case
data_dir = Path.home() / "repos/forobs/biotrade_data/"

# But you can override that with an environment variable
if os.environ.get("BIOTRADE_DATA"):
    data_dir = Path(os.environ["BIOTRADE_DATA"])

# Create if not existing
if not data_dir.exists():
    if os.environ.get("BIOTRADE_SKIP_CONFIRMATION"):
        data_dir.mkdir(parents=True)
    else:
        msg = f"Create {data_dir}?"
        if input(msg + "Please confirm [y/n]:") == "y":
            data_dir.mkdir(parents=True)
        else:
            print("Directory creation cancelled.")

# Database connection URL, default case
database_url = None
# But you can override it with an environment variable
if os.environ.get("BIOTRADE_DATABASE_URL"):
    database_url = os.environ["BIOTRADE_DATABASE_URL"]



"""
Written by Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

Script made to create statement of data policy, for the web platform

"""


def main():
    import pandas as pd
    import os
    import time
    from biotrade.faostat import faostat
    from scripts.front_end.functions import (
        save_file,
    )

    data_dir = faostat.data_dir
    info = os.stat(data_dir)
    # Folder modified
    db_update = time.strftime("%d-%m-%Y", time.localtime(info.st_mtime))
    statement_dict = {
        "source": ["PRDFAO", "TRDFAO", "TRDCOM"],
        "text_before": [
            "Data manipulation performed by the EU. Underlying data obtained from FAO. FAOSTAT Production and Forestry databases, licensed under ",
            "FAOSTAT: Data manipulation performed by the EU. Underlying data obtained from FAO. FAOSTAT Trade database, licensed under ",
            "UN COMTRADE: Data manipulation performed by the EU. Underlying data obtained from UNITED NATIONS, UN COMTRADE DATABASE: licensed under UN COMTRADE conditions available at ",
        ],
        "link1_target": [
            "https://creativecommons.org/licenses/by-nc-sa/3.0/igo/",
            "https://creativecommons.org/licenses/by-nc-sa/3.0/igo/",
            "https://shop.un.org/databases#Comtrade",
        ],
        "link1_text": [
            "CC-BY-NC-SA 3.0 IGO",
            "CC-BY-NC-SA 3.0 IGO",
            "https://shop.un.org/databases#Comtrade",
        ],
        "text_between": [
            ". Extracted from: ",
            ". Extracted from: ",
            ", extracted from ",
        ],
        "link2_target": [
            "https://www.fao.org/faostat/en/#home",
            "https://www.fao.org/faostat/en/#home",
            "https://comtradeplus.un.org/BulkFilesSearch",
        ],
        "link2_text": [
            "https://www.fao.org/faostat/en/#home",
            "https://www.fao.org/faostat/en/#home",
            "https://comtradeplus.un.org/BulkFilesSearch",
        ],
        "text_after": [
            f". Date of Access: {db_update}.",
            f". Date of Access: {db_update}.",
            f". Date of Access: {db_update}.",
        ],
    }
    df = pd.DataFrame(data=statement_dict)
    save_file(df, "statement.csv")


# Needed to avoid running module when imported
if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Test functions in :

    - faostat/mirror.py


"""

import numpy as np
import pandas
from biotrade.faostat.mirror import put_mirror_beside


def test_put_mirror_beside():
    df = pandas.DataFrame(
        {
            "reporter_code": [1, 2, 2],
            "partner_code": [2, 1, 3],
            "value": [10, 20, 30],
            "element": ["export", "import", "import"],
        }
    )
    output = put_mirror_beside(df)
    expected = pandas.DataFrame(
        {
            "reporter_code": [1, 2, 2, 3],
            "partner_code": [2, 1, 3, 2],
            "value": [10, 20, 30, np.nan],
            "element": ["export", "import", "import", "export"],
            "value_mirror": [20, 10, np.nan, 30],
        }
    )
    pandas.testing.assert_frame_equal(output, expected)



#!/usr/bin/env python
# coding: utf-8

""" Check that aggregates of forestry products correspond to the sum of their constituents

The purpose of this notebook is to compare aggregate products to the sum of
their constituents. Comparisons are done with the `biotrade.faostat.quality`
function `check_prod_agg_and_parts`.

Run this script at the command line with:

    ipython -i ~/repos/forobs/biotrade/scripts/quality/faostat_compare_aggregates_to_constituents.py

See also the notebook:

    ~/repos/forobs/obs3df_documents/notebooks/by_product/faostat_compare_aggregates_to_constituents.md

"""

# Internal modules
from biotrade.faostat.quality import check_prod_agg_and_parts
from biotrade.faostat import faostat


# Load FAOSTAT production data
fp = faostat.db.select("forestry_production")

#######################################################################################
# Compare production and trade of aggregate products to the sum of their constituents #
#######################################################################################

# ## Roundwood and sawnwood
# Check that roundwood production is the sum of industrial roundwood and fuel wood
# Also check for export and import values
check_prod_agg_and_parts(fp, "roundwood", ["industrial_roundwood", "wood_fuel"])

check_prod_agg_and_parts(
    fp, "sawnwood", ["sawnwood_coniferous", "sawnwood_non_coniferous_all"]
)

# ## Wood based panels
panel_constituents = [
    "plywood",
    "particle_board_and_osb_1961_1994_",
    "particle_board",
    "osb",
    "hardboard",
    "mdf_hdf",
    "other_fibreboard",
    "fibreboard_compressed_1961_1994_",
]
check_prod_agg_and_parts(
    fp.query("year> 1997"), "wood_based_panels", panel_constituents
)
# This will fail because some constituents only existed for recent years
# check_prod_agg_and_parts(
#     fp, "wood_based_panels", panel_constituents
# )

# ## Paper and paper board
paper_constituents = [
    "newsprint",
    "printing_and_writing_papers_uncoated_mechanical",
    "printing_and_writing_papers_uncoated_wood_free",
    "printing_and_writing_papers_coated",
    "household_and_sanitary_papers",
    "wrapping_and_packaging_paper_and_paperboard_1961_1997_",
    "case_materials",
    "cartonboard",
    "wrapping_papers",
    "other_papers_mainly_for_packaging",
    "other_paper_and_paperboard_n_e_s_not_elsewhere_specified_",
]
# Level 1 aggregate
# Constituents only exist for recent years
check_prod_agg_and_parts(
    fp.query("year > 1997"), "paper_and_paperboard", paper_constituents
)
# Constituents were not reported for old years, this will fail
check_prod_agg_and_parts(fp, "paper_and_paperboard", paper_constituents)

# Paper constituents level 2
paper_constituents_level_2 = [
    "graphic_papers",
    "other_paper_and_paperboard",
]
check_prod_agg_and_parts(fp, "paper_and_paperboard", paper_constituents_level_2)

# Printing and writing paper constituents
pwp_constituents = [
    "printing_and_writing_papers_uncoated_mechanical",
    "printing_and_writing_papers_uncoated_wood_free",
    "printing_and_writing_papers_coated",
]
# Constituents only exist for recent years
check_prod_agg_and_parts(
    fp.query("year > 1997"), "printing_and_writing_papers", pwp_constituents
)

# Constituents were not reported for old years, therefore this will fail
# check_prod_agg_and_parts(
#     fp, 'printing_and_writing_papers', pwp_constituents
# )

# Pulp
# Check that roundwood production is the sum of industrial roundwood and fuel wood
# Also check for export and import values
# Other aggregates
#    "pulp_for_paper",
#    "wood_pulp",
#     "wood_pulp_excluding_mechanical_wood_pulp",
#     "chemical_wood_pulp",
# Products not in wood pulp
#     "pulp_from_fibres_other_than_wood",
#     "recovered_fibre_pulp",
pulp_constituents = [
    "mechanical_and_semi_chemical_wood_pulp",
    "mechanical_wood_pulp",
    "semi_chemical_wood_pulp",
    "chemical_wood_pulp_sulphate_unbleached",
    "chemical_wood_pulp_sulphate_bleached",
    "chemical_wood_pulp_sulphite",
    "chemical_wood_pulp_sulphite_unbleached",
    "chemical_wood_pulp_sulphite_bleached",
    "dissolving_wood_pulp",
]
check_prod_agg_and_parts(fp, "wood_pulp", pulp_constituents)



"""
Written by Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

Script made to export data related to trade quantities of countries associated to the regulation and commodity tree products, for the web platform

"""


def main():
    from scripts.front_end.functions import (
        main_product_list,
        comtrade_products,
        merge_faostat_comtrade_data,
        filter_trade_data,
        aggregated_data,
        reporter_iso_codes,
        replace_zero_with_nan_values,
        save_file,
    )
    from biotrade.faostat import faostat
    from biotrade.faostat.aggregate import agg_trade_eu_row

    # Obtain faostat product codes
    faostat_list = main_product_list(["crop_trade"])
    # Obtain comtrade regulation codes
    comtrade_regulation = comtrade_products()
    # Trade data related to product code list
    trade_data = merge_faostat_comtrade_data(
        faostat_list, comtrade_regulation, aggregate=False
    )
    # Filter and change units
    trade_data = filter_trade_data(trade_data)
    # Aggregate french territories values to France and add them to the dataframe
    code_list = [68, 69, 87, 135, 182, 270, 281]
    agg_country_code = 68
    agg_country_name = faostat.country_groups.df
    agg_country_name = agg_country_name[
        agg_country_name.faost_code == agg_country_code
    ].fao_table_name.values[0]
    trade_data = aggregated_data(
        trade_data, code_list, agg_country_code, agg_country_name
    )
    # Substitute faostat codes with iso3 codes
    trade_data = reporter_iso_codes(trade_data)
    # Columns to be retained
    column_list = [
        "reporter_code",
        "partner_code",
        "product_code",
        "period",
        "value",
        "unit",
    ]
    # Consider selected columns of import quantities and values and save the files (drop nan)
    dropna_col = ["value"]
    trade_data = replace_zero_with_nan_values(trade_data, dropna_col)
    trade_data = trade_data.dropna(subset=dropna_col)
    df = trade_data[
        (trade_data["source"] == "faostat")
        & (trade_data["element"] == "import_quantity")
    ][column_list]
    save_file(df, "faostat_annual_variation.csv")
    df = trade_data[
        (trade_data["source"] == "faostat") & (trade_data["element"] == "import_value")
    ][column_list]
    save_file(df, "faostat_value_annual_variation.csv")
    # # Consider selected columns of export quantities and values and save the files (drop nan)
    # df = trade_data[
    #     (trade_data["source"] == "faostat")
    #     & (trade_data["element"] == "export_quantity")
    # ][column_list]
    # save_file(df, "faostat_annual_variation_mf.csv")
    # df = trade_data[
    #     (trade_data["source"] == "faostat") & (trade_data["element"] == "export_value")
    # ][column_list]
    # save_file(df, "faostat_value_annual_variation_mf.csv")
    # Consider selected columns of import quantities and values and save the files (drop nan)
    df = trade_data[
        (trade_data["source"] == "comtrade")
        & (trade_data["element"] == "import_quantity")
    ][column_list]
    save_file(df, "comtrade_annual_variation.csv")
    df = trade_data[
        (trade_data["source"] == "comtrade") & (trade_data["element"] == "import_value")
    ][column_list]
    save_file(df, "comtrade_value_annual_variation.csv")
    # # Consider selected columns of export quantities and values and save the files (drop nan)
    # df = trade_data[
    #     (trade_data["source"] == "comtrade")
    #     & (trade_data["element"] == "export_quantity")
    # ][column_list]
    # save_file(df, "comtrade_annual_variation_mf.csv")
    # df = trade_data[
    #     (trade_data["source"] == "comtrade") & (trade_data["element"] == "export_value")
    # ][column_list]
    # save_file(df, "comtrade_value_annual_variation_mf.csv")
    # Aggregate to EU and ROW for reporters
    eu_row_data = agg_trade_eu_row(
        trade_data,
        grouping_side="reporter",
        drop_index_col=[
            "year",
            "flag",
        ],
    )
    # Substitute with name and codes of the aggregations for the web platform
    selector = eu_row_data["reporter"] == "eu"
    eu_row_data.loc[selector, "reporter_code"] = "EU27"
    eu_row_data.loc[~selector, "reporter_code"] = "ROW"
    eu_row_data["reporter"].replace(
        ["eu", "row"], ["European Union", "Rest Of the World"], inplace=True
    )
    # Save imports
    df = eu_row_data[
        (eu_row_data["source"] == "faostat")
        & (eu_row_data["element"] == "import_quantity")
    ][column_list]
    save_file(df, "faostat_annual_variation_eu_row.csv")
    df = eu_row_data[
        (eu_row_data["source"] == "faostat")
        & (eu_row_data["element"] == "import_value")
    ][column_list]
    save_file(df, "faostat_value_annual_variation_eu_row.csv")
    df = eu_row_data[
        (eu_row_data["source"] == "comtrade")
        & (eu_row_data["element"] == "import_quantity")
    ][column_list]
    save_file(df, "comtrade_annual_variation_eu_row.csv")
    df = eu_row_data[
        (eu_row_data["source"] == "comtrade")
        & (eu_row_data["element"] == "import_value")
    ][column_list]
    save_file(df, "comtrade_value_annual_variation_eu_row.csv")
    # # Aggregate to EU and ROW for partners
    # eu_row_data = agg_trade_eu_row(
    #     trade_data,
    #     grouping_side="partner",
    #     drop_index_col=[
    #         "year",
    #         "flag",
    #     ],
    # )
    # # Substitute with name and codes of the aggregations for the web platform
    # selector = eu_row_data["partner"] == "eu"
    # eu_row_data.loc[selector, "partner_code"] = "EU27"
    # eu_row_data.loc[~selector, "partner_code"] = "ROW"
    # eu_row_data["partner"].replace(
    #     ["eu", "row"], ["European Union", "Rest Of the World"], inplace=True
    # )
    # # Save exports
    # df = eu_row_data[
    #     (eu_row_data["source"] == "faostat")
    #     & (eu_row_data["element"] == "export_quantity")
    # ][column_list]
    # save_file(df, "faostat_annual_variation_eu_row_mf.csv")
    # df = eu_row_data[
    #     (eu_row_data["source"] == "faostat")
    #     & (eu_row_data["element"] == "export_value")
    # ][column_list]
    # save_file(df, "faostat_value_annual_variation_eu_row_mf.csv")
    # df = eu_row_data[
    #     (eu_row_data["source"] == "comtrade")
    #     & (eu_row_data["element"] == "export_quantity")
    # ][column_list]
    # save_file(df, "comtrade_annual_variation_eu_row_mf.csv")
    # df = eu_row_data[
    #     (eu_row_data["source"] == "comtrade")
    #     & (eu_row_data["element"] == "export_value")
    # ][column_list]
    # save_file(df, "comtrade_value_annual_variation_eu_row_mf.csv")


# Needed to avoid running module when imported
if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Functions to analyse mirror flows.
"""
import logging

logger = logging.getLogger("biotrade.faostat")


def put_mirror_beside(df, drop_index_col=None):
    """Merge mirror flows to obtain a value column and a value_mirror column
    for the corresponding mirror flow.

    :param data frame df: bilateral trade flows in faostat compatible format,
        with an element column and a value column.
    :param drop_index_col list or str: variables to be dropped from the grouping index
    :return data frame with the same number of rows a new column value_mirror.

    The input data frame should contain data for both the export flow reported
    by the exporting country and the corresponding import flow reported by the
    importing country. The method swaps the reporter and partner columns (as
    well as reporter code and partner code), renames the value and flag columns
    to value_mirror and flag_mirror. Then merges back with the input data
    frame.

    Usage:

        >>> from biotrade.faostat import faostat
        >>> from biotrade.faostat.mirror import put_mirror_beside
        >>> swd = faostat.db.select(table="forestry_trade", product="sawnwood")
        >>> swd_mir = put_mirror_beside(swd)

    Show mirror values that are reported as exactly the same

        >>> selector = swd_mir["value"] == swd_mir["value_mirror"]
        >>> swd_mir[selector]

    Show mirror values that are different or that are empty

        >>> swd_mir[~selector & ~swd_mir["value_mirror"].isna()]
        >>> swd_mir[swd_mir["value_mirror"].isna()]

    Select crop trade where products contain the word "soy" and then add the
    value_mirror column.

        >>> soy = faostat.db.select(table="crop_trade", product = "soy")
        >>> soy_mir = put_mirror_beside(soy)

    Get mirror values for trade flows aggregated at EU and ROW level

        >>> import pandas
        >>> from biotrade.faostat import faostat
        >>> from biotrade.faostat.mirror import put_mirror_beside
        >>> from biotrade.faostat.aggregate import agg_trade_eu_row
        >>> country = "Ukraine"
        >>> # Load trade reported by the country and group EU-ROW on the partner side
        >>> dfr = faostat.db.select("crop_trade", reporter=country, product_code="268")
        >>> dfragg = agg_trade_eu_row(dfr, grouping_side="partner")
        >>> # Load trade reported by its partners and group EU-ROW on the reporter side
        >>> dfp = faostat.db.select("crop_trade", partner=country, product_code="268")
        >>> dfpagg = agg_trade_eu_row(dfp, grouping_side="reporter")
        >>> # Concatenate the 2 datasets and put mirror value beside the value
        >>> dfm = put_mirror_beside(pandas.concat([dfragg, dfpagg]))

    """
    # Give drop_index_col its default value and make it a list
    if drop_index_col is None:
        drop_index_col = ["flag"]
    if isinstance(drop_index_col, str):
        drop_index_col = [drop_index_col]
    # Swap reporter and partner columns
    df_m = df.copy()
    reporter_cols = ["reporter", "reporter_code"]
    reporter_cols = [x for x in reporter_cols if x in df.columns]
    partner_cols = ["partner", "partner_code"]
    partner_cols = [x for x in partner_cols if x in df.columns]
    df_m[reporter_cols] = df[partner_cols]
    df_m[partner_cols] = df[reporter_cols]
    # Swap element names
    df_m["element"] = df_m["element"].str.replace("import", "xxx")
    df_m["element"] = df_m["element"].str.replace("export", "import")
    df_m["element"] = df_m["element"].str.replace("xxx", "export")
    # Drop the element_code column
    if "element_code" in df.columns:
        df_m.drop(columns="element_code", inplace=True)
    # Build the aggregation index based on all columns in df_m
    # Removing the "value" and the drop_index_col columns
    index = df_m.columns.to_list()
    for col in drop_index_col + ["value"]:
        if col in df.columns:
            index.remove(col)
    logger.info("\nMerging mirror flows on the following index:\n %s", index)
    # Merge with the original data frame
    return df.merge(df_m, on=index, how="outer", suffixes=("", "_mirror"))



"""
Written by Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

JRC biomass Project.
Unit D1 Bioeconomy.

"""

# Internal modules
from biotrade.comtrade import comtrade

# Third party modules
from sqlalchemy.sql import func
from sqlalchemy import select
import pandas as pd


def compare_yearly_agg_monthly():
    """
    Function which compares Comtrade yearly data with respect to the aggregation of monthly data.
    It returns a dataframe containing aggregation values which don't match

    """

    # Table with yearly data to select from raw_comtrade schema
    table = comtrade.db.tables["yearly"]
    # Left table columns to be considered before performing the join
    columns = [
        table.c.reporter_code,
        table.c.reporter,
        table.c.partner_code,
        table.c.partner,
        table.c.product_code,
        table.c.flow_code,
        table.c.flow,
        table.c.period,
        table.c.year,
    ]
    # Add quantity and monetary value columns
    table_stmt = table.select().with_only_columns(
        [
            *columns,
            table.c.net_weight.label("qnt_yearly"),
            table.c.trade_value.label("value_yearly"),
        ]
    )
    # Consider years after 2015
    table_stmt = table_stmt.where(table.c.year >= 2016)
    # Define the new table name
    yearly_selection = table_stmt.subquery().alias("yearly")
    # Table with monthly data to select from raw_comtrade schema
    table = comtrade.db.tables["monthly"]
    # Right table Columns to be considered before performing the join
    columns = [
        table.c.reporter_code,
        table.c.reporter,
        table.c.partner_code,
        table.c.partner,
        table.c.product_code,
        table.c.flow_code,
        table.c.flow,
        table.c.year,
    ]
    # Add aggregation quantity and monetary value columns
    table_stmt = table.select().with_only_columns(
        [
            *columns,
            func.sum(table.c.net_weight).label("qnt_monthly"),
            func.sum(table.c.trade_value).label("value_monthly"),
        ]
    )
    # Consider years after 2015
    table_stmt = table_stmt.where(table.c.year >= 2016)
    # Perform aggregation
    table_stmt = table_stmt.group_by(*columns)
    # Define the new table name
    monthly_selection = table_stmt.subquery().alias("monthly")
    # Full outer join to calculate the difference between results
    table_join = yearly_selection.join(
        monthly_selection,
        (yearly_selection.c.reporter_code == monthly_selection.c.reporter_code)
        & (yearly_selection.c.partner_code == monthly_selection.c.partner_code)
        & (yearly_selection.c.product_code == monthly_selection.c.product_code)
        & (yearly_selection.c.flow_code == monthly_selection.c.flow_code)
        & (yearly_selection.c.year == monthly_selection.c.year),
        isouter=True,
        full=True,
    )
    table = select(table_join).alias("table_join")
    # Final columns to be retained
    join_columns = [
        table.c.reporter_code,
        table.c.reporter,
        table.c.partner_code,
        table.c.partner,
        table.c.product_code,
        table.c.flow_code,
        table.c.flow,
        table.c.period,
        table.c.year,
        table.c.qnt_yearly,
        table.c.qnt_monthly,
        func.abs(table.c.qnt_yearly - table.c.qnt_monthly).label("abs_qnt_diff"),
        table.c.value_yearly,
        table.c.value_monthly,
        func.abs(table.c.value_yearly - table.c.value_monthly).label("abs_value_diff"),
    ]
    table_stmt = select(join_columns)
    # Return data where monthly aggregation quantities or monetary values don't match with yearly data
    table_stmt = table_stmt.where(
        (table.c.qnt_yearly != table.c.qnt_monthly)
        | ((table.c.qnt_yearly == None) & (table.c.qnt_monthly != None))
        | ((table.c.qnt_yearly != None) & (table.c.qnt_monthly == None))
        | (table.c.value_yearly != table.c.value_monthly)
        | ((table.c.value_yearly == None) & (table.c.value_monthly != None))
        | ((table.c.value_yearly != None) & (table.c.value_monthly == None))
    )
    # Return the dataframe from the query to db
    df = pd.read_sql_query(table_stmt, comtrade.db.engine)
    return df


def compare_world_agg_country(table):
    """
    Function which compares Comtrade data with partner equal to World
    and country partner aggregations. It returns a dataframe containing aggregation values which don't match with World data

    """

    # Table to select from raw_comtrade schema
    table = comtrade.db.tables[table]
    # Columns to be considered before performing the join
    columns = [
        table.c.reporter_code,
        table.c.reporter,
        table.c.product_code,
        table.c.flow_code,
        table.c.flow,
        table.c.period,
        table.c.year,
    ]
    # Add aggregation quantity and monetary value columns
    table_stmt = table.select().with_only_columns(
        [
            *columns,
            func.sum(table.c.net_weight).label("qnt_agg"),
            func.sum(table.c.trade_value).label("value_agg"),
        ]
    )
    # Consider all partners except World (partner_code = 0)
    table_stmt = table_stmt.where(table.c.partner_code > 0)
    # Perform aggregation
    table_stmt = table_stmt.group_by(*columns)
    # Define the new table name
    agg_selection = table_stmt.subquery().alias("agg")
    # Perform the same operations for partner_code = 0 (World)
    table_stmt = table.select().with_only_columns(
        [
            *columns,
            func.sum(table.c.net_weight).label("qnt_world"),
            func.sum(table.c.trade_value).label("value_world"),
        ]
    )
    table_stmt = table_stmt.where(table.c.partner_code == 0)
    table_stmt = table_stmt.group_by(*columns)
    world_selection = table_stmt.subquery().alias("world")
    # Full outer join to calculate the difference between results
    table_join = agg_selection.join(
        world_selection,
        (agg_selection.c.reporter_code == world_selection.c.reporter_code)
        & (agg_selection.c.product_code == world_selection.c.product_code)
        & (agg_selection.c.flow_code == world_selection.c.flow_code)
        & (agg_selection.c.period == world_selection.c.period),
        isouter=True,
        full=True,
    )
    table = select(table_join).alias("table_join")
    # Final columns to be retained
    join_columns = [
        table.c.reporter_code,
        table.c.reporter,
        table.c.product_code,
        table.c.flow_code,
        table.c.flow,
        table.c.period,
        table.c.year,
        table.c.qnt_agg,
        table.c.qnt_world,
        func.abs(table.c.qnt_agg - table.c.qnt_world).label("abs_qnt_diff"),
        table.c.value_agg,
        table.c.value_world,
        func.abs(table.c.value_agg - table.c.value_world).label("abs_value_diff"),
    ]
    table_stmt = select(join_columns)
    # Return data where aggregation quantities or monetary values don't match with World data provided by Comtrade
    table_stmt = table_stmt.where(
        (table.c.qnt_agg != table.c.qnt_world)
        | ((table.c.qnt_agg == None) & (table.c.qnt_world != None))
        | ((table.c.qnt_agg != None) & (table.c.qnt_world == None))
        | (table.c.value_agg != table.c.value_world)
        | ((table.c.value_agg == None) & (table.c.value_world != None))
        | ((table.c.value_agg != None) & (table.c.value_world == None))
    )
    # Return the dataframe from the query to db
    df = pd.read_sql_query(table_stmt, comtrade.db.engine)
    return df



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Test the import reallocation functions

"""

import pandas
from pandas.testing import assert_series_equal
from pandas.testing import assert_frame_equal

from biotrade.common.reallocate import allocate_by_partners
from biotrade.common.reallocate import compute_share_by_partners
from biotrade.common.reallocate import compute_share_prod_imp
from biotrade.common.reallocate import split_prod_imp


def test_compute_share_prod_imp():
    df_prod = pandas.DataFrame(
        {
            "year": 1,
            "reporter": ["a", "b"],
            "reporter_code": [1, 2],
            "product": ["p", "p"],
            "production": [1, 2],
        }
    )
    df_trade = pandas.DataFrame(
        {
            "year": 1,
            "reporter": ["a", "a", "b", "b"],
            "reporter_code": [1, 1, 2, 2],
            "partner": ["x", "y", "x", "z"],
            "partner_code": [24, 25, 24, 26],
            "product": ["p", "p", "p", "p"],
            "import_quantity": [1, 2, 3, 4],
        }
    )
    output = compute_share_prod_imp(df_prod, df_trade)
    expected = pandas.Series([0.25, 2 / 9], name="share_prod_imp")
    assert_series_equal(output["share_prod_imp"], expected)


def test_split_prod_imp():
    df = pandas.DataFrame(
        {
            "reporter": ["a", "b", "c"],
            "reporter_code": [1, 2, 3],
            "primary_product": ["p", "p", "q"],
            "year": 1,
            "primary_eq_0": [1, 2, 4],
        }
    )
    df_share = pandas.DataFrame(
        {
            "reporter": ["a", "b", "c"],
            "reporter_code": [1, 2, 3],
            "primary_product": ["p", "p", "q"],
            "year": 1,
            "share_prod_imp": [0, 1, 0.5],
        }
    )
    df_output = split_prod_imp(df, df_share, 1)
    # Use float in the expected series to avoid AssertionError:
    # Attributes of Series are different Attribute "dtype" are different
    # [left]: float64 [right]: int64
    expected_prod = pandas.Series([0, 2, 2.0])
    expected_imp = pandas.Series([1, 0, 2.0])
    assert_series_equal(
        df_output["primary_eq_prod_1"], expected_prod, check_names=False
    )
    assert_series_equal(df_output["primary_eq_imp_1"], expected_imp, check_names=False)


def test_allocate_by_partners():
    df_prod = pandas.DataFrame(
        {
            "year": 1,
            "reporter": ["a", "b"],
            "reporter_code": [1, 2],
            "primary_product": ["p", "p"],
            "primary_eq_imp_1": [6, 14],
        }
    )
    df_trade = pandas.DataFrame(
        {
            "year": 1,
            "reporter": ["a", "a", "b", "b"],
            "reporter_code": [1, 1, 2, 2],
            "partner": ["x", "y", "x", "z"],
            "partner_code": [24, 25, 24, 26],
            "primary_product": ["p", "p", "p", "p"],
            "import_quantity": [1, 2, 3, 4],
        }
    )
    expected = df_trade.copy()
    expected.rename(columns={"partner": "partner_1"}, inplace=True)
    expected = expected[["year", "reporter", "partner_1", "primary_product"]]
    # Use float in the expected series
    expected["primary_eq_imp_alloc_1"] = [2, 4, 6, 8.0]
    # Share by partners
    df_trade_2 = compute_share_by_partners(df_trade, threshold=0)
    output = allocate_by_partners(df_prod, df_trade_2, 1)
    assert_frame_equal(output[expected.columns], expected)


# In [31]: oil production columns
# Index(['reporter_code', 'reporter', 'product_code', 'product', 'element_code',
#        'element', 'period', 'year', 'unit', 'value', 'flag',
#        'extraction_rate_mean', 'extraction_rate', 'oil_value_share',
#        'primary_eq'],
#       dtype='object')
#
# In [32]:  trade columns
# Out[32]: Index(['reporter', 'partner', 'product', 'period', 'value'], dtype='object')



"""
Written by Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

Script made to export data for the EUFO web platform, related to lafo values of countries associated to the regulation and commodity tree products

"""


def main():
    from scripts.front_end.functions import (
        aggregated_data,
        reporter_iso_codes,
        replace_zero_with_nan_values,
        save_file,
    )
    from biotrade.faostat import faostat
    from biotrade.comtrade import comtrade
    from biotrade.faostat.aggregate import agg_trade_eu_row
    import pandas as pd

    # TODO retrieve lafo data from functions (deforestation package), for now reading a csv example file
    lafo_data = pd.read_csv(comtrade.data_dir / "lafo_example.csv")
    # Render columns lower case
    lafo_data.columns = lafo_data.columns.str.lower()
    # Remove columns
    lafo_data.drop(
        columns=[
            "iso2_code",
            "primary_commodity_code",
            "primary_commodity",
            "area_harvested",
            "production",
            "yield",
            "yield_avg",
            "flag_out",
        ],
        inplace=True,
    )
    # Rename to product_code column
    lafo_data.rename(columns={"reg_code": "product_code"}, inplace=True)
    # Add period column
    lafo_data["period"] = lafo_data["year"]
    # Aggregate french territories values to France and add them to the dataframe
    code_list = [68, 69, 87, 135, 182, 270, 281]
    agg_country_code = 68
    agg_country_name = faostat.country_groups.df
    agg_country_name = agg_country_name[
        agg_country_name.faost_code == agg_country_code
    ].fao_table_name.values[0]
    lafo_data = aggregated_data(
        lafo_data,
        code_list,
        agg_country_code,
        agg_country_name,
        groupby_cols=[
            "product_code",
            "year",
        ],
        value_cols=["pce", "lafo_obs"],
    )
    # Substitute faostat codes with iso3 codes
    lafo_data = reporter_iso_codes(lafo_data)
    # Add units
    lafo_data["pce_unit"] = "tonnes"
    lafo_data["lafo_obs_unit"] = "ha"
    # Aggregate to EU and ROW for reporters
    eu_row_data = agg_trade_eu_row(
        lafo_data,
        grouping_side="reporter",
        drop_index_col=[
            "year",
        ],
        value_col=["pce", "lafo_obs"],
    )
    # Substitute with name and codes of the aggregations for the web platform
    selector = eu_row_data["reporter"] == "eu"
    eu_row_data.loc[selector, "reporter_code"] = "EU27"
    eu_row_data.loc[~selector, "reporter_code"] = "ROW"
    eu_row_data["reporter"].replace(
        ["eu", "row"], ["European Union", "Rest Of the World"], inplace=True
    )
    # Columns to be finally retained
    column_list = [
        "reporter_code",
        "partner_code",
        "product_code",
        "period",
        "value",
        "unit",
    ]
    # Consider selected columns of import quantities and values and save the files (drop nan)
    dropna_col = ["value"]
    for dataset in [lafo_data, eu_row_data]:
        for value_col in ["pce", "lafo_obs"]:
            for source in ["faostat", "comtrade"]:
                df = dataset.copy()
                df.rename(
                    columns={value_col: "value", (value_col + "_unit"): "unit"},
                    inplace=True,
                )
                df = replace_zero_with_nan_values(df, dropna_col)
                df = df.dropna(subset=dropna_col)
                # Faostat
                file_name = f"{source}_{value_col.split('_', 1)[0]}_annual_variation"
                if dataset.equals(eu_row_data):
                    file_name = file_name + "_eu_row"
                save_file(
                    df[df["source"] == source][column_list],
                    file_name + ".csv",
                )


# Needed to avoid running module when imported
if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Paul Rougieux and Selene Patani.

Copyright (c) 2023 European Union
Licenced under the MIT licence

JRC biomass Project.
Unit D1 Bioeconomy.

You can use this object at the ipython console with the following examples.

Download only and return a data frame, for debugging purposes:

    >>> from biotrade.comtrade import comtrade
    >>> # Other sawnwood
    >>> swd99 = comtrade.pump.download(cc = "440799")
    >>> # Soy
    >>> soy = comtrade.pump.download(cc = "120190")

Download and store in the database as used when updating the database

    >>> comtrade.pump.append_to_db(cc = "440799")

Show the list of reporter countries

    >>> from biotrade.comtrade import comtrade
    >>> comtrade.country_groups.reporters

Display information on column names used for renaming
and dropping less important columns:

    >>> comtrade.column_names

"""
# Third party modules
import pandas

# Internal modules
from biotrade import module_dir, data_dir, database_url
from biotrade.comtrade.country_groups import CountryGroups
from biotrade.comtrade.database import DatabaseComtradePostgresql
from biotrade.comtrade.database import DatabaseComtradeSqlite
from biotrade.comtrade.dump import Dump
from biotrade.comtrade.products import Products
from biotrade.comtrade.pump import Pump

# Define a logging mechanism to keep track of errors and debug messages
from biotrade.common.logger import create_logger

create_logger()


class Comtrade:
    """
    Parent to various objects dealing with data from the UN Comtrade API.
    """

    # Location of module configuration data
    config_data_dir = module_dir / "config_data"

    # Load a mapping table used to rename columns
    df = pandas.read_csv(config_data_dir / "column_names.csv")
    # Select only relevant columns and remove incomplete mappings
    df = df[
        ["biotrade", "comtrade_machine", "comtrade_human", "comtrade_apicall"]
    ]
    column_names = df

    def __init__(self):
        # Location of the data
        self.data_dir = data_dir / "comtrade"
        if not self.data_dir.exists():
            self.data_dir.mkdir()

    @property
    def country_groups(self):
        """Identify reporter and partner countries and regions"""
        return CountryGroups(self)

    @property
    def db_pgsql(self):
        """Store Comtrade data and make it available for further processing"""
        return DatabaseComtradePostgresql(self)

    @property
    def db_sqlite(self):
        """Store Comtrade data and make it available for further processing"""
        return DatabaseComtradeSqlite(self)

    @property
    def db(self):
        """The generic database can be either a PostGreSQL or a SQLite database
        Depending of the value of the DATABASE_URL variable. If it's None
        then use SQLite otherwise use the PostGreSQL db defined in the
        environmental variable BIOTRADE_DATABASE_URL.
        """
        if database_url is None:
            return DatabaseComtradeSqlite(self)
        return DatabaseComtradePostgresql(self)

    @property
    def dump(self):
        """Dump data from the database into files"""
        return Dump(self)

    @property
    def products(self):
        """Identify Comtrade products (commodities) and metadata"""
        return Products(self)

    @property
    def pump(self):
        """Download data from Comtrade and send it to the database"""
        return Pump(self)


# Make a singleton #
comtrade = Comtrade()
